import logging
from io import SEEK_CUR
from typing import Iterable, Iterator

from .exceptions import FlvDataError
from .format import FlvDumper, FlvParser
from .io_protocols import RandomIO
from .models import BACK_POINTER_SIZE, FlvHeader, FlvTag
from .utils import AutoRollbacker, OffsetRepositor

__all__ = 'FlvReader', 'FlvWriter'


logger = logging.getLogger(__name__)


class FlvReader:
    def __init__(
        self,
        stream: RandomIO,
        *,
        backup_timestamp: bool = False,
        restore_timestamp: bool = False,
    ) -> None:
        self._stream = stream
        self._backup_timestamp = backup_timestamp
        self._restore_timestamp = restore_timestamp
        self._parser = FlvParser(
            stream,
            backup_timestamp=backup_timestamp,
            restore_timestamp=restore_timestamp,
        )

    def read_header(self) -> FlvHeader:
        header = self._parser.parse_header()
        previous_tag_size = self._parser.parse_previous_tag_size()
        if previous_tag_size != 0:
            raise FlvDataError(f'First back-pointer must be 0: {previous_tag_size}')
        return header

    def read_tag(self, *, no_body: bool = False) -> FlvTag:
        tag = self._parser.parse_tag(no_body=no_body)
        previous_tag_size = self._parser.parse_previous_tag_size()
        if previous_tag_size != tag.tag_size:
            raise FlvDataError(f'Wrong back-pointer: {previous_tag_size}')
        return tag

    def read_tags(self, *, no_body: bool = False) -> Iterator[FlvTag]:
        while True:
            try:
                yield self.read_tag(no_body=no_body)
            except EOFError:
                break

    def rread_tag(self, *, no_body: bool = False) -> FlvTag:
        self._seek_to_previous_tag()
        with OffsetRepositor(self._stream):
            return self.read_tag(no_body=no_body)

    def rread_tags(self, *, no_body: bool = False) -> Iterator[FlvTag]:
        while True:
            try:
                yield self.rread_tag(no_body=no_body)
            except EOFError:
                break

    def read_body(self, tag: FlvTag) -> bytes:
        with OffsetRepositor(self._stream):
            self._stream.seek(tag.body_offset)
            return self._stream.read(tag.body_size)

    def _seek_to_previous_tag(self) -> int:
        try:
            self._stream.seek(-BACK_POINTER_SIZE, SEEK_CUR)
        except OSError as e:
            if e.errno == 22 and e.strerror == 'Invalid argument':
                raise EOFError()
            else:
                raise
        previous_tag_size = self._parser.parse_previous_tag_size()
        return self._stream.seek(-(4 + previous_tag_size), SEEK_CUR)


class FlvWriter:
    def __init__(self, stream: RandomIO) -> None:
        self._stream = stream
        self._dumper = FlvDumper(stream)

    def write_header(self, header: FlvHeader) -> int:
        with AutoRollbacker(self._stream):
            self._dumper.dump_header(header)
            self._dumper.dump_previous_tag_size(0)
            return header.size + BACK_POINTER_SIZE

    def write_tag(self, tag: FlvTag) -> int:
        with AutoRollbacker(self._stream):
            self._dumper.dump_tag(tag)
            self._dumper.dump_previous_tag_size(tag.tag_size)
            return tag.tag_size + BACK_POINTER_SIZE

    def write_tags(self, tags: Iterable[FlvTag]) -> int:
        return sum(map(self.write_tag, tags))



import atexit
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as _TimeoutError
from typing import Callable, Any, Iterable, Mapping, TypeVar


_T = TypeVar('_T')


_executor = None


def wait_for(
    func: Callable[..., _T],
    *,
    args: Iterable[Any] = [],
    kwargs: Mapping[str, Any] = {},
    timeout: float
) -> _T:
    global _executor
    if _executor is None:
        _executor = ThreadPoolExecutor(thread_name_prefix='wait_for')
        atexit.register(_executor.shutdown)

    future = _executor.submit(func, *args, **kwargs)
    try:
        return future.result(timeout=timeout)
    except _TimeoutError:
        raise TimeoutError(timeout, func, args, kwargs) from None



import time
from typing import Callable, Iterator, Optional, TypeVar

from reactivex import Observable, abc, catch_with_iterable
from reactivex import operators as ops

_T = TypeVar('_T')


def retry(
    count: Optional[int] = None,
    delay: Optional[float] = None,
    should_retry: Callable[[Exception], bool] = lambda _: True,
) -> Callable[[Observable[_T]], Observable[_T]]:
    def _retry(source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            exception: Optional[Exception] = None

            def counter() -> Iterator[int]:
                n = 0
                while True:
                    if exception is not None:
                        if not should_retry(exception):
                            break
                        if count is not None and n > count:
                            break
                        if delay:
                            time.sleep(delay)
                    yield n
                    n += 1

            def on_error(e: Exception) -> None:
                nonlocal exception
                exception = e

            _source = source.pipe(ops.do_action(on_error=on_error))
            return catch_with_iterable(_source for _ in counter()).subscribe(
                observer, scheduler=scheduler
            )

        return Observable(subscribe)

    return _retry



import errno
import io
import logging
from contextlib import suppress
import os
import re
import shlex
from subprocess import PIPE, CalledProcessError, Popen
from threading import Condition, Thread
from typing import Optional, cast

from ..utils.io import wait_for
from ..utils.mixins import StoppableMixin, SupportDebugMixin

logger = logging.getLogger(__name__)


__all__ = ('StreamRemuxer',)


class FFmpegError(Exception):
    pass


class StreamRemuxer(StoppableMixin, SupportDebugMixin):
    _ERROR_PATTERN = re.compile(
        r'\b(error|failed|missing|invalid|corrupt)\b', re.IGNORECASE
    )

    def __init__(self, room_id: int, remove_filler_data: bool = False) -> None:
        super().__init__()
        self._room_id = room_id
        self._remove_filler_data = remove_filler_data
        self._exception: Optional[Exception] = None
        self._ready = Condition()
        self._env = None

        self._init_for_debug(room_id)
        if self._debug:
            self._env = os.environ.copy()
            path = os.path.join(self._debug_dir, f'ffreport-{room_id}-%t.log')
            self._env['FFREPORT'] = f'file={path}:level=48'

    @property
    def input(self) -> io.BufferedWriter:
        assert self._subprocess.stdin is not None
        return cast(io.BufferedWriter, self._subprocess.stdin)

    @property
    def output(self) -> io.BufferedReader:
        assert self._subprocess.stdout is not None
        return cast(io.BufferedReader, self._subprocess.stdout)

    @property
    def exception(self) -> Optional[Exception]:
        return self._exception

    def __enter__(self):  # type: ignore
        self.start()
        self.wait()
        return self

    def __exit__(self, exc_type, value, traceback):  # type: ignore
        self.stop()
        self.raise_for_exception()

    def wait(self, timeout: Optional[float] = None) -> bool:
        with self._ready:
            return self._ready.wait(timeout=timeout)

    def restart(self) -> None:
        logger.debug('Restarting stream remuxer...')
        self.stop()
        self.start()
        logger.debug('Restarted stream remuxer')

    def raise_for_exception(self) -> None:
        if not self.exception:
            return
        raise self.exception

    def _do_start(self) -> None:
        logger.debug('Starting stream remuxer...')
        self._thread = Thread(
            target=self._run, name=f'StreamRemuxer::{self._room_id}', daemon=True
        )
        self._thread.start()

    def _do_stop(self) -> None:
        logger.debug('Stopping stream remuxer...')
        if hasattr(self, '_subprocess'):
            with suppress(ProcessLookupError):
                self._subprocess.kill()
            self._subprocess.wait(timeout=10)
        if hasattr(self, '_thread'):
            self._thread.join(timeout=10)

    def _run(self) -> None:
        logger.debug('Started stream remuxer')
        self._exception = None
        try:
            self._run_subprocess()
        except BrokenPipeError as exc:
            logger.debug(repr(exc))
        except FFmpegError as exc:
            if not self._stopped:
                logger.warning(repr(exc))
            else:
                logger.debug(repr(exc))
        except TimeoutError as exc:
            logger.debug(repr(exc))
        except Exception as exc:
            # OSError: [Errno 22] Invalid argument
            # https://stackoverflow.com/questions/23688492/oserror-errno-22-invalid-argument-in-subprocess
            if isinstance(exc, OSError) and exc.errno == errno.EINVAL:
                pass
            else:
                self._exception = exc
                logger.exception(exc)
        finally:
            self._stopped = True
            logger.debug('Stopped stream remuxer')

    def _run_subprocess(self) -> None:
        cmd = 'ffmpeg -xerror -i pipe:0 -c copy -copyts'
        if self._remove_filler_data:
            cmd += ' -bsf:v filter_units=remove_types=12'
        cmd += ' -f flv pipe:1'
        args = shlex.split(cmd)

        with Popen(
            args, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=self._env
        ) as self._subprocess:
            with self._ready:
                self._ready.notify_all()

            assert self._subprocess.stderr is not None
            with io.TextIOWrapper(
                self._subprocess.stderr, encoding='utf-8', errors='backslashreplace'
            ) as stderr:
                while not self._stopped:
                    line = wait_for(stderr.readline, timeout=10)
                    if not line:
                        if self._subprocess.poll() is not None:
                            break
                        else:
                            continue
                    if self._debug:
                        logger.debug('ffmpeg: %s', line)
                    self._check_error(line)

        if not self._stopped and self._subprocess.returncode not in (0, 255):
            # 255: Exiting normally, received signal 2.
            raise CalledProcessError(self._subprocess.returncode, cmd=cmd)

    def _check_error(self, line: str) -> None:
        match = self._ERROR_PATTERN.search(line)
        if not match:
            return
        raise FFmpegError(line)



from __future__ import annotations

import io
import logging
from typing import List, Optional, cast

from reactivex import Observable, Subject, abc

from ...utils.ffprobe import StreamProfile, ffprobe
from ..common import find_aac_header_tag, find_avc_header_tag
from ..io import FlvWriter
from ..models import FlvHeader, FlvTag
from .typing import FLVStream, FLVStreamItem

__all__ = ('Prober', 'StreamProfile')


logger = logging.getLogger(__name__)


class Prober:
    def __init__(self) -> None:
        self._profiles: Subject[StreamProfile] = Subject()

    def _reset(self) -> None:
        self._gathering: bool = False
        self._gathered_items: List[FLVStreamItem] = []

    @property
    def profiles(self) -> Observable[StreamProfile]:
        return self._profiles

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._probe(source)

    def _probe(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            self._reset()

            def on_next(item: FLVStreamItem) -> None:
                if isinstance(item, FlvHeader):
                    self._gathered_items.clear()
                    self._gathering = True

                if self._gathering:
                    self._gathered_items.append(item)
                    if len(self._gathered_items) >= 10:
                        try:
                            self._do_probe()
                        except Exception as e:
                            logger.warning(f'Failed to probe stream: {repr(e)}')
                        finally:
                            self._gathered_items.clear()
                            self._gathering = False

                observer.on_next(item)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _do_probe(self) -> None:
        bytes_io = io.BytesIO()
        writer = FlvWriter(bytes_io)
        writer.write_header(cast(FlvHeader, self._gathered_items[0]))
        gathered_tags = cast(List[FlvTag], self._gathered_items[1:])

        avc_sequence_header = find_avc_header_tag(gathered_tags)
        aac_sequence_header = find_aac_header_tag(gathered_tags)
        if avc_sequence_header is not None and aac_sequence_header is not None:
            index_of_avc_sequence_header = gathered_tags.index(avc_sequence_header)
            index_of_aac_sequence_header = gathered_tags.index(aac_sequence_header)
            if index_of_avc_sequence_header > index_of_aac_sequence_header:
                gathered_tags[index_of_aac_sequence_header] = avc_sequence_header
                gathered_tags[index_of_avc_sequence_header] = aac_sequence_header

        for tag in gathered_tags:
            writer.write_tag(tag)

        def on_next(profile: StreamProfile) -> None:
            self._profiles.on_next(profile)

        def on_error(e: Exception) -> None:
            logger.warning(f'Failed to probe stream by ffprobe: {repr(e)}')

        ffprobe(bytes_io.getvalue()).subscribe(on_next, on_error)



from __future__ import annotations

from typing import Any, Dict, Iterable, Iterator, Mapping, Optional, Union

import attr
from typing_extensions import TypeGuard

from . import scriptdata
from .avc import extract_resolution
from .io import FlvReader
from .io_protocols import RandomIO
from .models import AudioTag, AVCPacketType, FlvTag, ScriptTag, TagType, VideoTag
from .utils import OffsetRepositor


def read_tags(
    reader: FlvReader, count: int, *, no_body: bool = False
) -> Iterator[FlvTag]:
    assert count > 0, 'count must greater than 0'
    for c, tag in enumerate(reader.read_tags(no_body=no_body), start=1):
        yield tag
        if c >= count:
            break


def rread_tags(
    reader: FlvReader, count: int, *, no_body: bool = False
) -> Iterator[FlvTag]:
    assert count > 0, 'count must greater than 0'
    for c, tag in enumerate(reader.rread_tags(no_body=no_body), start=1):
        yield tag
        if c >= count:
            break


def read_tags_in_duration(
    reader: FlvReader, duration: int, *, no_body: bool = False
) -> Iterator[FlvTag]:
    for tag in reader.read_tags(no_body=no_body):
        yield tag
        if tag.timestamp > 0:
            break
    else:
        raise EOFError('no tags')

    start = tag.timestamp
    end = start + duration

    for tag in reader.read_tags(no_body=no_body):
        yield tag
        if tag.timestamp >= end:
            break


def peek_tags(
    file: RandomIO, reader: FlvReader, count: int, *, no_body: bool = False
) -> Iterator[FlvTag]:
    with OffsetRepositor(file):
        yield from read_tags(reader, count, no_body=no_body)


def rpeek_tags(
    file: RandomIO, reader: FlvReader, count: int, *, no_body: bool = False
) -> Iterator[FlvTag]:
    with OffsetRepositor(file):
        yield from rread_tags(reader, count, no_body=no_body)


def find_metadata_tag(tags: Iterable[FlvTag]) -> Optional[ScriptTag]:
    for tag in tags:
        if is_script_tag(tag):
            return tag
    return None


def find_header_tag(tags: Iterable[FlvTag]) -> Optional[Union[AudioTag, VideoTag]]:
    for tag in tags:
        if is_sequence_header(tag):
            return tag
    return None


def find_avc_header_tag(tags: Iterable[FlvTag]) -> Optional[VideoTag]:
    for tag in tags:
        if is_video_sequence_header(tag):
            return tag
    return None


def find_aac_header_tag(tags: Iterable[FlvTag]) -> Optional[AudioTag]:
    for tag in tags:
        if is_audio_sequence_header(tag):
            return tag
    return None


def find_nalu_keyframe_tag(tags: Iterable[FlvTag]) -> Optional[VideoTag]:
    for tag in tags:
        if is_video_nalu_keyframe(tag):
            return tag
    return None


def find_aac_raw_tag(tags: Iterable[FlvTag]) -> Optional[AudioTag]:
    for tag in tags:
        if is_audio_tag(tag) and tag.is_aac_raw():
            return tag
    return None


def is_audio_tag(tag: FlvTag) -> TypeGuard[AudioTag]:
    return tag.tag_type == TagType.AUDIO


def is_video_tag(tag: FlvTag) -> TypeGuard[VideoTag]:
    return tag.tag_type == TagType.VIDEO


def is_script_tag(tag: FlvTag) -> TypeGuard[ScriptTag]:
    return tag.tag_type == TagType.SCRIPT


def is_metadata_tag(tag: FlvTag) -> TypeGuard[ScriptTag]:
    if is_script_tag(tag):
        script_data = parse_scriptdata(tag)
        return script_data['name'] == 'onMetaData'
    return False


def is_data_tag(tag: FlvTag) -> TypeGuard[Union[AudioTag, VideoTag]]:
    return is_audio_data_tag(tag) or is_video_data_tag(tag)


def is_audio_data_tag(tag: FlvTag) -> TypeGuard[AudioTag]:
    return is_audio_tag(tag) and tag.is_aac_raw()


def is_video_data_tag(tag: FlvTag) -> TypeGuard[VideoTag]:
    return is_video_tag(tag) and tag.is_avc_nalu()


def is_sequence_header(tag: FlvTag) -> TypeGuard[Union[AudioTag, VideoTag]]:
    return is_audio_sequence_header(tag) or is_video_sequence_header(tag)


def is_audio_sequence_header(tag: FlvTag) -> TypeGuard[AudioTag]:
    return is_audio_tag(tag) and tag.is_aac_header()


def is_video_sequence_header(tag: FlvTag) -> TypeGuard[VideoTag]:
    return is_video_tag(tag) and tag.is_avc_header()


def is_video_nalu_keyframe(tag: FlvTag) -> TypeGuard[VideoTag]:
    return is_video_tag(tag) and tag.is_keyframe() and tag.is_avc_nalu()


def parse_scriptdata(script_tag: ScriptTag) -> scriptdata.ScriptData:
    assert script_tag.body is not None
    return scriptdata.load(script_tag.body)


def unparse_scriptdata(
    script_tag: ScriptTag, script_data: scriptdata.ScriptData, **changes: Any
) -> ScriptTag:
    body = scriptdata.dump(script_data)
    return script_tag.evolve(body=body, **changes)


def create_script_tag(
    script_data: scriptdata.ScriptData, offset: int = 0, timestamp: int = 0
) -> ScriptTag:
    body = scriptdata.dump(script_data)
    return ScriptTag(
        filtered=False,
        tag_type=TagType.SCRIPT,
        data_size=len(body),
        timestamp=timestamp,
        stream_id=0,
        offset=offset,
        body=body,
    )


def create_metadata_tag(
    metadata: Dict[str, Any], offset: int = 0, timestamp: int = 0
) -> ScriptTag:
    script_data = scriptdata.ScriptData(name='onMetaData', value=metadata)
    return create_script_tag(script_data, offset, timestamp)


def parse_metadata(metadata_tag: ScriptTag) -> Dict[str, Any]:
    script_data = parse_scriptdata(metadata_tag)
    assert script_data['name'] == 'onMetaData'
    return script_data['value']


def unparse_metadata(
    metadata_tag: ScriptTag, metadata: Dict[str, Any], **changes: Any
) -> ScriptTag:
    script_data = scriptdata.ScriptData(name='onMetaData', value=metadata)
    return unparse_scriptdata(metadata_tag, script_data, **changes)


def enrich_metadata(
    metadata_tag: ScriptTag,
    extra_metadata: Mapping[str, Any],
    offset: Optional[int] = None,
) -> ScriptTag:
    metadata = parse_metadata(metadata_tag)
    metadata.update(extra_metadata)
    metadata = ensure_order(metadata)

    if offset is None:
        return unparse_metadata(metadata_tag, metadata)
    else:
        return unparse_metadata(metadata_tag, metadata, offset=offset)


def update_metadata(
    metadata_tag: ScriptTag, metadata: Mapping[str, Any], offset: Optional[int] = None
) -> ScriptTag:
    original_tag_size = metadata_tag.tag_size
    new_tag = enrich_metadata(metadata_tag, metadata, offset)
    assert new_tag.tag_size == original_tag_size, 'tag size must unchanged'
    return new_tag


def ensure_order(metadata: Dict[str, Any]) -> Dict[str, Any]:
    # the order of metadata property matters!!!
    # some typical properties such as 'keyframes' must be before some custom
    # properties such as 'Comment' otherwise, it won't take effect in some
    # players!
    from .operators import MetaData

    typical_props = attr.fields_dict(MetaData).keys()
    return {
        **{k: v for k, v in metadata.items() if k in typical_props},
        **{k: v for k, v in metadata.items() if k not in typical_props},
    }


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Resolution:
    width: int
    height: int

    @classmethod
    def from_metadata(cls, metadata: Dict[str, Any]) -> Resolution:
        return cls(
            width=metadata.get('width') or metadata['displayWidth'],
            height=metadata.get('height') or metadata['displayHeight'],
        )

    @classmethod
    def from_aac_sequence_header(cls, tag: VideoTag) -> Resolution:
        assert tag.avc_packet_type == AVCPacketType.AVC_SEQUENCE_HEADER
        assert tag.body is not None
        width, height = extract_resolution(tag.body)
        return cls(width, height)



import logging
import asyncio
from typing import Any, Dict, List

import aiohttp
from tenacity import (
    AsyncRetrying,
    wait_exponential,
    stop_after_delay,
)

from .models import WebHook
from ..utils.mixins import SwitchableMixin
from ..exception import ExceptionCenter
from ..event import EventCenter, Error, ErrorData
from ..event.typing import Event
from .. import __prog__, __version__


__all__ = 'WebHookEmitter',


logger = logging.getLogger(__name__)


class WebHookEmitter(SwitchableMixin):
    def __init__(
        self, webhooks: List[WebHook] = []
    ) -> None:
        super().__init__()
        self.webhooks = webhooks
        self.headers = {
            'User-Agent': f'{__prog__}/{__version__}'
        }

    def _do_enable(self) -> None:
        events = EventCenter.get_instance().events
        self._event_subscription = events.subscribe(self._on_event)
        exceptions = ExceptionCenter.get_instance().exceptions
        self._exc_subscription = exceptions.subscribe(self._on_exception)

    def _do_disable(self) -> None:
        self._event_subscription.dispose()
        self._exc_subscription.dispose()

    def _on_event(self, event: Event) -> None:
        for webhook in self.webhooks:
            if isinstance(event, webhook.event_types):
                self._send_event(webhook.url, event)

    def _on_exception(self, exc: BaseException) -> None:
        for webhook in self.webhooks:
            if webhook.receive_exception:
                self._send_exception(webhook.url, exc)

    def _send_event(self, url: str, event: Event) -> None:
        self._send_request(url, event.asdict())

    def _send_exception(self, url: str, exc: BaseException) -> None:
        payload = Error.from_data(ErrorData.from_exc(exc)).asdict()
        self._send_request(url, payload)

    def _send_request(self, url: str, payload: Dict[str, Any]) -> None:
        asyncio.create_task(self._send_request_async(url, payload))

    async def _send_request_async(
        self, url: str, payload: Dict[str, Any]
    ) -> None:
        try:
            async for attempt in AsyncRetrying(
                stop=stop_after_delay(180),
                wait=wait_exponential(max=15),
            ):
                with attempt:
                    await self._post(url, payload)
        except Exception as e:
            logger.warning('Failed to send a request to {}: {}'.format(
                url, repr(e)
            ))

    async def _post(self, url: str, payload: Dict[str, Any]) -> None:
        async with aiohttp.ClientSession(
            headers=self.headers,
            raise_for_status=True,
        ) as session:
            async with session.post(url, json=payload):
                pass



from __future__ import annotations

import io
import logging
from typing import Optional

import requests
from reactivex import Observable, abc

from ...bili.live import Live
from ...utils.mixins import AsyncCooperationMixin

__all__ = ('StreamFetcher',)


logger = logging.getLogger(__name__)


class StreamFetcher(AsyncCooperationMixin):
    def __init__(
        self,
        live: Live,
        session: requests.Session,
        *,
        read_timeout: Optional[int] = None,
    ) -> None:
        super().__init__()
        self._live = live
        self._session = session
        self.read_timeout = read_timeout or 3

    def __call__(self, source: Observable[str]) -> Observable[io.RawIOBase]:
        return self._fetch(source)

    def _fetch(self, source: Observable[str]) -> Observable[io.RawIOBase]:
        def subscribe(
            observer: abc.ObserverBase[io.RawIOBase],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(url: str) -> None:
                try:
                    logger.info(f'Requesting live stream... {url}')
                    response = self._session.get(
                        url,
                        stream=True,
                        headers=self._live.headers,
                        timeout=self.read_timeout,
                    )
                    logger.info('Response received')
                    response.raise_for_status()
                except Exception as e:
                    logger.warning(f'Failed to request live stream: {repr(e)}')
                    observer.on_error(e)
                else:
                    observer.on_next(response.raw)  # urllib3.response.HTTPResponse

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)





from .io_protocols import RandomIO


def format_timestamp(ts: int) -> str:
    milliseconds = ts % 1000
    seconds = ts // 1000 % 60
    minutes = ts // 1000 // 60 % 60
    hours = ts // 1000 // 60 // 60 % 60
    return f'{hours:0>2}:{minutes:0>2}:{seconds:0>2}.{milliseconds:0>3}'


def format_offest(offset: int) -> str:
    return f'0x{offset:0>8x}'


class OffsetRepositor:
    def __init__(self, file: RandomIO):
        self._file = file

    def __enter__(self):  # type: ignore
        self._offset = self._file.tell()

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        self._file.seek(self._offset)


class AutoRollbacker:
    def __init__(self, file: RandomIO):
        self._file = file

    def __enter__(self):  # type: ignore
        self._offset = self._file.tell()

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        if exc_type is not None:
            self._file.seek(self._offset)
            self._file.truncate()



from __future__ import annotations
from enum import Enum
from typing import Optional

import attr

from ..bili.models import RoomInfo, UserInfo
from ..bili.typing import StreamFormat, QualityNumber
from ..core.cover_downloader import CoverSaveStrategy
from ..postprocess import DeleteStrategy, PostprocessorStatus
from ..postprocess.typing import Progress


class RunningStatus(str, Enum):
    STOPPED = 'stopped'
    WAITING = 'waiting'
    RECORDING = 'recording'
    REMUXING = 'remuxing'
    INJECTING = 'injecting'


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TaskStatus:
    monitor_enabled: bool
    recorder_enabled: bool
    running_status: RunningStatus
    stream_url: str
    stream_host: str
    dl_total: int  # Number of Bytes in total
    dl_rate: float  # Number of Bytes per second
    rec_elapsed: float  # time elapsed
    rec_total: int  # Number of Bytes in total
    rec_rate: float  # Number of Bytes per second
    danmu_total: int  # Number of Danmu in total
    danmu_rate: float  # Number of Danmu per minutes
    real_stream_format: Optional[StreamFormat]
    real_quality_number: Optional[QualityNumber]
    recording_path: Optional[str] = None
    postprocessor_status: PostprocessorStatus = PostprocessorStatus.WAITING
    postprocessing_path: Optional[str] = None
    postprocessing_progress: Optional[Progress] = None


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TaskParam:
    # OutputSettings
    out_dir: str
    path_template: str
    filesize_limit: int
    duration_limit: int
    # HeaderSettings
    user_agent: str
    cookie: str
    # DanmakuSettings
    danmu_uname: bool
    record_gift_send: bool
    record_free_gifts: bool
    record_guard_buy: bool
    record_super_chat: bool
    save_raw_danmaku: bool
    # RecorderSettings
    stream_format: StreamFormat
    quality_number: QualityNumber
    fmp4_stream_timeout: int
    read_timeout: int
    disconnection_timeout: Optional[int]
    buffer_size: int
    save_cover: bool
    cover_save_strategy: CoverSaveStrategy
    # PostprocessingOptions
    remux_to_mp4: bool
    inject_extra_metadata: bool
    delete_source: DeleteStrategy


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TaskData:
    user_info: UserInfo
    room_info: RoomInfo
    task_status: TaskStatus


class VideoFileStatus(str, Enum):
    RECORDING = 'recording'
    REMUXING = 'remuxing'
    INJECTING = 'injecting'
    COMPLETED = 'completed'
    MISSING = 'missing'
    BROKEN = 'broken'


class DanmukuFileStatus(str, Enum):
    RECORDING = 'recording'
    COMPLETED = 'completed'
    MISSING = 'missing'
    BROKEN = 'broken'


@attr.s(auto_attribs=True, slots=True, frozen=True)
class VideoFileDetail:
    path: str
    size: int
    status: VideoFileStatus


@attr.s(auto_attribs=True, slots=True, frozen=True)
class DanmakuFileDetail:
    path: str
    size: int
    status: DanmukuFileStatus



import logging
from typing import Callable, Optional

from reactivex import Observable, abc

from ..common import is_audio_sequence_header, is_metadata_tag, is_video_sequence_header
from ..models import AudioTag, FlvHeader, ScriptTag, VideoTag
from .correct import correct
from .typing import FLVStream, FLVStreamItem

__all__ = ('split',)

logger = logging.getLogger(__name__)


def split() -> Callable[[FLVStream], FLVStream]:
    def _split(source: FLVStream) -> FLVStream:
        """Split the FLV stream when av parameters are changed."""

        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            changed: bool = False
            last_flv_header: Optional[FlvHeader] = None
            last_metadata_tag: Optional[ScriptTag] = None
            last_audio_sequence_header: Optional[AudioTag] = None
            last_video_sequence_header: Optional[VideoTag] = None

            def reset() -> None:
                nonlocal changed
                nonlocal last_flv_header
                nonlocal last_metadata_tag
                nonlocal last_audio_sequence_header, last_video_sequence_header
                changed = False
                last_flv_header = None
                last_metadata_tag = None
                last_audio_sequence_header = last_video_sequence_header = None

            def insert_header_and_tags() -> None:
                assert last_flv_header is not None
                observer.on_next(last_flv_header)
                if last_metadata_tag is not None:
                    observer.on_next(last_metadata_tag)
                if last_video_sequence_header is not None:
                    observer.on_next(last_video_sequence_header)
                if last_audio_sequence_header is not None:
                    observer.on_next(last_audio_sequence_header)

            def on_next(item: FLVStreamItem) -> None:
                nonlocal changed
                nonlocal last_flv_header
                nonlocal last_metadata_tag
                nonlocal last_audio_sequence_header, last_video_sequence_header

                if isinstance(item, FlvHeader):
                    reset()
                    last_flv_header = item
                    observer.on_next(item)
                    return

                tag = item

                if is_metadata_tag(tag):
                    logger.debug(f'Metadata tag: {tag}')
                    last_metadata_tag = tag
                elif is_audio_sequence_header(tag):
                    logger.debug(f'Audio sequence header: {tag}')
                    if last_audio_sequence_header is not None:
                        if not tag.is_the_same_as(last_audio_sequence_header):
                            logger.warning('Audio parameters changed')
                            changed = True
                        last_audio_sequence_header = tag
                        return
                    last_audio_sequence_header = tag
                elif is_video_sequence_header(tag):
                    logger.debug(f'Video sequence header: {tag}')
                    if last_video_sequence_header is not None:
                        if not tag.is_the_same_as(last_video_sequence_header):
                            logger.warning('Video parameters changed')
                            changed = True
                        last_video_sequence_header = tag
                        return
                    last_video_sequence_header = tag
                else:
                    if changed:
                        logger.debug('Splitting stream...')
                        changed = False
                        insert_header_and_tags()
                        logger.debug('Splitted stream')

                observer.on_next(tag)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe).pipe(correct())

    return _split



from __future__ import annotations

import logging
import time
from datetime import datetime
from typing import Optional

import m3u8
import requests
import urllib3
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential

from ...bili.live import Live
from ...utils.mixins import SupportDebugMixin

__all__ = ('PlaylistFetcher',)


logger = logging.getLogger(__name__)


class PlaylistFetcher(SupportDebugMixin):
    def __init__(self, live: Live, session: requests.Session) -> None:
        super().__init__()
        self._init_for_debug(live.room_id)
        self._live = live
        self._session = session

    def __call__(self, source: Observable[str]) -> Observable[m3u8.M3U8]:
        return self._fetch(source)

    def _fetch(self, source: Observable[str]) -> Observable[m3u8.M3U8]:
        def subscribe(
            observer: abc.ObserverBase[m3u8.M3U8],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            if self._debug:
                path = '{}/playlist-{}-{}.m3u8'.format(
                    self._debug_dir,
                    self._live.room_id,
                    datetime.now().strftime('%Y-%m-%d-%H%M%S-%f'),
                )
                playlist_debug_file = open(path, 'wt', encoding='utf-8')

            disposed = False
            subscription = SerialDisposable()

            def on_next(url: str) -> None:
                logger.info(f'Fetching playlist... {url}')
                while not disposed:
                    try:
                        content = self._fetch_playlist(url)
                        if self._debug:
                            playlist_debug_file.write(content + '\n')
                        playlist = m3u8.loads(content, uri=url)
                        if playlist.is_variant:
                            url = self._get_best_quality_url(playlist)
                            logger.debug('Playlist changed to variant playlist')
                            on_next(url)
                    except Exception as e:
                        logger.warning(f'Failed to fetch playlist: {repr(e)}')
                        observer.on_error(e)
                    else:
                        observer.on_next(playlist)
                        time.sleep(1)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                if self._debug:
                    playlist_debug_file.close()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _get_best_quality_url(self, playlist: m3u8.M3U8) -> str:
        sorted_playlists = sorted(
            playlist.playlists, key=lambda p: p.stream_info.bandwidth
        )
        return sorted_playlists[-1].absolute_uri

    @retry(
        reraise=True,
        retry=retry_if_exception_type(
            (
                requests.exceptions.Timeout,
                urllib3.exceptions.TimeoutError,
                urllib3.exceptions.ProtocolError,
            )
        ),
        wait=wait_exponential(multiplier=0.1, max=1),
        stop=stop_after_delay(10),
    )
    def _fetch_playlist(self, url: str) -> str:
        response = self._session.get(url, headers=self._live.headers, timeout=3)
        response.raise_for_status()
        response.encoding = 'utf-8'
        return response.text



from typing import Union


from .models import (
    LiveBeganEvent,
    LiveBeganEventData,
    LiveEndedEvent,
    LiveEndedEventData,
    RoomChangeEvent,
    RoomChangeEventData,
    RecordingStartedEvent,
    RecordingStartedEventData,
    RecordingFinishedEvent,
    RecordingFinishedEventData,
    RecordingCancelledEvent,
    RecordingCancelledEventData,
    VideoFileCreatedEvent,
    VideoFileCreatedEventData,
    VideoFileCompletedEvent,
    VideoFileCompletedEventData,
    DanmakuFileCreatedEvent,
    DanmakuFileCreatedEventData,
    DanmakuFileCompletedEvent,
    DanmakuFileCompletedEventData,
    RawDanmakuFileCreatedEvent,
    RawDanmakuFileCreatedEventData,
    RawDanmakuFileCompletedEvent,
    RawDanmakuFileCompletedEventData,
    VideoPostprocessingCompletedEvent,
    VideoPostprocessingCompletedEventData,
    SpaceNoEnoughEvent,
    SpaceNoEnoughEventData,
    Error,
    ErrorData,
)


Event = Union[
    LiveBeganEvent,
    LiveEndedEvent,
    RoomChangeEvent,
    RecordingStartedEvent,
    RecordingFinishedEvent,
    RecordingCancelledEvent,
    VideoFileCreatedEvent,
    VideoFileCompletedEvent,
    DanmakuFileCreatedEvent,
    DanmakuFileCompletedEvent,
    RawDanmakuFileCreatedEvent,
    RawDanmakuFileCompletedEvent,
    VideoPostprocessingCompletedEvent,
    SpaceNoEnoughEvent,
    Error,
]

EventData = Union[
    LiveBeganEventData,
    LiveEndedEventData,
    RoomChangeEventData,
    RecordingStartedEventData,
    RecordingFinishedEventData,
    RecordingCancelledEventData,
    VideoFileCreatedEventData,
    VideoFileCompletedEventData,
    DanmakuFileCreatedEventData,
    DanmakuFileCompletedEventData,
    RawDanmakuFileCreatedEventData,
    RawDanmakuFileCompletedEventData,
    VideoPostprocessingCompletedEventData,
    SpaceNoEnoughEventData,
    ErrorData,
]



from __future__ import annotations

import io
from typing import Any, Optional

from reactivex import Observable, Subject, abc

from ..statistics import Statistics

__all__ = ('StreamStatistics',)


class StreamStatistics:
    def __init__(self) -> None:
        self._statistics = Statistics()

    @property
    def count(self) -> int:
        return self._statistics.count

    @property
    def rate(self) -> float:
        return self._statistics.rate

    @property
    def elapsed(self) -> float:
        return self._statistics.elapsed

    def freeze(self) -> None:
        self._statistics.freeze()

    def reset(self) -> None:
        self._statistics.reset()

    def __call__(self, source: Observable[io.RawIOBase]) -> Observable[io.RawIOBase]:
        self.reset()
        return self._calc(source)

    def _calc(self, source: Observable[io.RawIOBase]) -> Observable[io.RawIOBase]:
        def subscribe(
            observer: abc.ObserverBase[io.RawIOBase],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(stream: io.RawIOBase) -> None:
                calculable_stream = CalculableStream(stream)
                calculable_stream.size_updates.subscribe(self._statistics.submit)
                observer.on_next(calculable_stream)

            def on_completed() -> None:
                self._statistics.freeze()
                observer.on_completed()

            return source.subscribe(
                on_next, observer.on_error, on_completed, scheduler=scheduler
            )

        return Observable(subscribe)


class CalculableStream(io.RawIOBase):
    def __init__(self, stream: io.RawIOBase) -> None:
        self._stream = stream
        self._offset: int = 0
        self._size_updates: Subject[int] = Subject()

    @property
    def size_updates(self) -> Observable[int]:
        return self._size_updates

    def read(self, size: int = -1) -> bytes:
        data = self._stream.read(size)
        assert data is not None
        self._offset += len(data)
        self._size_updates.on_next(len(data))
        return data

    def readinto(self, b: Any) -> int:
        n = self._stream.readinto(b)
        assert n is not None
        self._offset += n
        self._size_updates.on_next(n)
        return n

    def tell(self) -> int:
        return self._offset




from enum import Enum


class PostprocessorStatus(Enum):
    WAITING = 'waiting'
    REMUXING = 'remuxing'
    INJECTING = 'injecting'

    def __str__(self) -> str:
        return self.value


class DeleteStrategy(Enum):
    AUTO = 'auto'
    SAFE = 'safe'
    NEVER = 'never'

    def __str__(self) -> str:
        return self.value

    # workaround for value serialization
    def __repr__(self) -> str:
        return str(self)



from abc import ABC
import hashlib
from urllib.parse import urlencode
from datetime import datetime
from typing import Mapping, Dict, Any, Final

import aiohttp
from tenacity import (
    retry,
    wait_exponential,
    stop_after_delay,
)

from .typing import QualityNumber, JsonResponse, ResponseData
from .exceptions import ApiRequestError


__all__ = 'AppApi', 'WebApi'


class BaseApi(ABC):
    def __init__(self, session: aiohttp.ClientSession):
        self._session = session
        self.timeout = 10

    @staticmethod
    def _check_response(json_res: JsonResponse) -> None:
        if json_res['code'] != 0:
            raise ApiRequestError(
                json_res['code'],
                json_res.get('message') or json_res.get('msg') or '',
            )

    @retry(
        reraise=True,
        stop=stop_after_delay(5),
        wait=wait_exponential(0.1),
    )
    async def _get(self, *args: Any, **kwds: Any) -> JsonResponse:
        async with self._session.get(
            *args,
            **kwds,
            timeout=self.timeout,
        ) as res:
            json_res = await res.json()
            self._check_response(json_res)
            return json_res


class AppApi(BaseApi):
    # taken from https://github.com/SocialSisterYi/bilibili-API-collect/blob/master/other/API_sign.md  # noqa
    _appkey = '1d8b6e7d45233436'
    _appsec = '560c52ccd288fed045859ed18bffd973'

    _headers = {
        'User-Agent': 'Mozilla/5.0 BiliDroid/6.64.0 (bbcallen@gmail.com) os/android model/Unknown mobi_app/android build/6640400 channel/bili innerVer/6640400 osVer/6.0.1 network/2',  # noqa
        'Connection': 'Keep-Alive',
        'Accept-Encoding': 'gzip',
    }

    @classmethod
    def signed(cls, params: Mapping[str, Any]) -> Dict[str, Any]:
        if isinstance(params, Mapping):
            params = dict(sorted({**params, 'appkey': cls._appkey}.items()))
        else:
            raise ValueError(type(params))
        query = urlencode(params, doseq=True)
        sign = hashlib.md5((query + cls._appsec).encode()).hexdigest()
        params.update(sign=sign)
        return params

    async def get_room_play_info(
        self,
        room_id: int,
        qn: QualityNumber = 10000,
        *,
        only_video: bool = False,
        only_audio: bool = False,
    ) -> ResponseData:
        url = 'https://api.live.bilibili.com/xlive/app-room/v2/index/getRoomPlayInfo'  # noqa

        params = self.signed({
            'actionKey': 'appkey',
            'build': '6640400',
            'channel': 'bili',
            'codec': '0,1',  # 0: avc, 1: hevc
            'device': 'android',
            'device_name': 'Unknown',
            'disable_rcmd': '0',
            'dolby': '1',
            'format': '0,1,2',  # 0: flv, 1: ts, 2: fmp4
            'free_type': '0',
            'http': '1',
            'mask': '0',
            'mobi_app': 'android',
            'need_hdr': '0',
            'no_playurl': '0',
            'only_audio': '1' if only_audio else '0',
            'only_video': '1' if only_video else '0',
            'platform': 'android',
            'play_type': '0',
            'protocol': '0,1',
            'qn': qn,
            'room_id': room_id,
            'ts': int(datetime.utcnow().timestamp()),
        })

        r = await self._get(url, params=params, headers=self._headers)
        return r['data']

    async def get_info_by_room(self, room_id: int) -> ResponseData:
        url = 'https://api.live.bilibili.com/xlive/app-room/v1/index/getInfoByRoom'  # noqa

        params = self.signed({
            'actionKey': 'appkey',
            'build': '6640400',
            'channel': 'bili',
            'device': 'android',
            'mobi_app': 'android',
            'platform': 'android',
            'room_id': room_id,
            'ts': int(datetime.utcnow().timestamp()),
        })

        r = await self._get(url, params=params)
        return r['data']

    async def get_user_info(self, uid: int) -> ResponseData:
        url = 'https://app.bilibili.com/x/v2/space'

        params = self.signed({
            'build': '6640400',
            'channel': 'bili',
            'mobi_app': 'android',
            'platform': 'android',
            'ts': int(datetime.utcnow().timestamp()),
            'vmid': uid,
        })

        r = await self._get(url, params=params)
        return r['data']

    async def get_danmu_info(self, room_id: int) -> ResponseData:
        url = 'https://api.live.bilibili.com/xlive/app-room/v1/index/getDanmuInfo'  # noqa

        params = self.signed({
            'actionKey': 'appkey',
            'build': '6640400',
            'channel': 'bili',
            'device': 'android',
            'mobi_app': 'android',
            'platform': 'android',
            'room_id': room_id,
            'ts': int(datetime.utcnow().timestamp()),
        })

        r = await self._get(url, params=params)
        return r['data']


class WebApi(BaseApi):
    BASE_API_URL: Final[str] = 'https://api.bilibili.com'
    BASE_LIVE_API_URL: Final[str] = 'https://api.live.bilibili.com'

    GET_USER_INFO_URL: Final[str] = BASE_API_URL + '/x/space/acc/info'

    GET_DANMU_INFO_URL: Final[str] = BASE_LIVE_API_URL + \
        '/xlive/web-room/v1/index/getDanmuInfo'
    ROOM_INIT_URL: Final[str] = BASE_LIVE_API_URL + '/room/v1/Room/room_init'
    GET_INFO_URL: Final[str] = BASE_LIVE_API_URL + '/room/v1/Room/get_info'
    GET_INFO_BY_ROOM_URL: Final[str] = BASE_LIVE_API_URL + \
        '/xlive/web-room/v1/index/getInfoByRoom'
    GET_ROOM_PLAY_INFO_URL: Final[str] = BASE_LIVE_API_URL + \
        '/xlive/web-room/v2/index/getRoomPlayInfo'
    GET_TIMESTAMP_URL: Final[str] = BASE_LIVE_API_URL + \
        '/av/v1/Time/getTimestamp?platform=pc'

    async def room_init(self, room_id: int) -> ResponseData:
        r = await self._get(self.ROOM_INIT_URL, params={'id': room_id})
        return r['data']

    async def get_room_play_info(
        self, room_id: int, qn: QualityNumber = 10000
    ) -> ResponseData:
        params = {
            'room_id': room_id,
            'protocol': '0,1',
            'format': '0,1,2',
            'codec': '0,1',
            'qn': qn,
            'platform': 'web',
            'ptype': 8,
        }
        r = await self._get(self.GET_ROOM_PLAY_INFO_URL, params=params)
        return r['data']

    async def get_info_by_room(self, room_id: int) -> ResponseData:
        params = {
            'room_id': room_id,
        }
        r = await self._get(self.GET_INFO_BY_ROOM_URL, params=params)
        return r['data']

    async def get_info(self, room_id: int) -> ResponseData:
        params = {
            'room_id': room_id,
        }
        r = await self._get(self.GET_INFO_URL, params=params)
        return r['data']

    async def get_timestamp(self) -> int:
        r = await self._get(self.GET_TIMESTAMP_URL)
        return r['data']['timestamp']

    async def get_user_info(self, uid: int) -> ResponseData:
        params = {
            'mid': uid,
        }
        r = await self._get(self.GET_USER_INFO_URL, params=params)
        return r['data']

    async def get_danmu_info(self, room_id: int) -> ResponseData:
        params = {
            'id': room_id,
        }
        r = await self._get(self.GET_DANMU_INFO_URL, params=params)
        return r['data']



import os
from typing import Callable, Optional, TypeVar

from reactivex import Observable, abc

_T = TypeVar('_T')


def replace(src_path: str, dst_path: str) -> Callable[[Observable[_T]], Observable[_T]]:
    def _replace(source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_completed() -> None:
                try:
                    os.replace(src_path, dst_path)
                except Exception as e:
                    observer.on_error(e)
                else:
                    observer.on_completed()

            return source.subscribe(
                observer.on_next, observer.on_error, on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    return _replace



from __future__ import annotations

import asyncio
import logging
from typing import Optional, TypeVar

import requests
import urllib3
from reactivex import Observable, abc

from ...utils import operators as utils_ops

__all__ = ('RequestExceptionHandler',)


logger = logging.getLogger(__name__)

_T = TypeVar('_T')


class RequestExceptionHandler:
    def __call__(self, source: Observable[_T]) -> Observable[_T]:
        return self._handle(source).pipe(
            utils_ops.retry(delay=1, should_retry=self._should_retry)
        )

    def _handle(self, source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_error(exc: Exception) -> None:
                try:
                    raise exc
                except asyncio.exceptions.TimeoutError:
                    logger.warning(repr(exc))
                except requests.exceptions.Timeout:
                    logger.warning(repr(exc))
                except requests.exceptions.HTTPError:
                    logger.warning(repr(exc))
                except urllib3.exceptions.TimeoutError:
                    logger.warning(repr(exc))
                except urllib3.exceptions.ProtocolError:
                    # ProtocolError('Connection broken: IncompleteRead(
                    logger.warning(repr(exc))
                except Exception:
                    pass
                observer.on_error(exc)

            return source.subscribe(
                observer.on_next, on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(
            exc,
            (
                asyncio.exceptions.TimeoutError,
                requests.exceptions.Timeout,
                requests.exceptions.HTTPError,
                urllib3.exceptions.TimeoutError,
                urllib3.exceptions.ProtocolError,
            ),
        ):
            return True
        else:
            return False



from __future__ import annotations

import logging
from enum import IntEnum, auto
from typing import Callable, List, Optional, TypedDict, cast

import attr
from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from typing_extensions import TypeGuard

from ...utils.hash import cksum
from ..common import (
    create_script_tag,
    is_audio_sequence_header,
    is_metadata_tag,
    is_script_tag,
    is_sequence_header,
    is_video_sequence_header,
    parse_scriptdata,
)
from ..models import AudioTag, FlvHeader, FlvTag, ScriptTag, VideoTag
from ..scriptdata import ScriptData
from ..utils import format_timestamp
from .typing import FLVStream, FLVStreamItem

__all__ = ('concat', 'JoinPointExtractor', 'JoinPoint', 'JoinPointData')

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class JoinPoint:
    seamless: bool
    timestamp: float  # timestamp of next tag in milliseconds
    crc32: str  # crc32 of the next tag

    @classmethod
    def from_metadata_value(cls, value: JoinPointData) -> JoinPoint:
        return cls(
            timestamp=int(value['timestamp']),
            seamless=value['seamless'],
            crc32=value['crc32'],
        )

    def to_metadata_value(self) -> JoinPointData:
        return dict(
            timestamp=float(self.timestamp), seamless=self.seamless, crc32=self.crc32
        )

    def __str__(self) -> str:
        return 'seamless: {}, timestamp: {}, crc32: {}'.format(
            'yes' if self.seamless else 'no',
            format_timestamp(int(self.timestamp)),
            self.crc32,
        )


class JoinPointData(TypedDict):
    seamless: bool
    timestamp: float
    crc32: str


class ACTION(IntEnum):
    NOOP = auto()
    CORRECT = auto()
    GATHER = auto()
    CANCEL = auto()
    CONCAT = auto()
    CONCAT_AND_GATHER = auto()


def concat(
    num_of_last_tags: int = 3, max_duration: int = 20_000
) -> Callable[[FLVStream], FLVStream]:
    """Concat FLV streams.

    num_of_last_tags: Number of tags for determining whether or not tags are duplicated
    max_duration: Max duration in milliseconds the duplicated tags might last
    """

    def _concat(source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            delta: int = 0
            action: ACTION = ACTION.NOOP
            last_tags: List[FlvTag] = []
            gathered_tags: List[FlvTag] = []
            last_flv_header: Optional[FlvHeader] = None
            last_audio_sequence_header: Optional[AudioTag] = None
            last_video_sequence_header: Optional[VideoTag] = None

            def update_last_tags(tag: FlvTag) -> None:
                nonlocal last_audio_sequence_header, last_video_sequence_header
                last_tags.append(tag)
                if len(last_tags) > num_of_last_tags:
                    last_tags.pop(0)
                if is_audio_sequence_header(tag):
                    last_audio_sequence_header = tag
                elif is_video_sequence_header(tag):
                    last_video_sequence_header = tag

            def gather_tags(tag: FlvTag) -> None:
                nonlocal gathered_tags
                nonlocal action
                nonlocal last_audio_sequence_header, last_video_sequence_header
                if is_audio_sequence_header(tag):
                    if last_audio_sequence_header is None:
                        logger.debug(
                            'Cancel concat due to no last audio sequence header'
                        )
                        action = ACTION.CANCEL
                    else:
                        if not tag.is_the_same_as(last_audio_sequence_header):
                            action = ACTION.CANCEL
                            logger.debug(
                                'Cancel concat due to audio sequence header changed'
                            )
                    last_audio_sequence_header = tag
                elif is_video_sequence_header(tag):
                    if last_video_sequence_header is None:
                        logger.debug(
                            'Cancel concat due to no last video sequence header'
                        )
                        action = ACTION.CANCEL
                    else:
                        if not tag.is_the_same_as(last_video_sequence_header):
                            action = ACTION.CANCEL
                            logger.debug(
                                'Cancel concat due to video sequence header changed'
                            )
                    last_video_sequence_header = tag
                gathered_tags.append(tag)

            def has_gathering_completed() -> bool:
                # XXX: timestamp MUST start from 0 and continuous!
                # put the correct and fix operator on upstream of this operator
                # to ensure timestamp start from 0 and continuous!
                return gathered_tags[-1].timestamp >= max_duration

            def find_last_duplicated_tag(tags: List[FlvTag]) -> int:
                logger.debug('Finding duplicated tags...')

                last_out_tag = last_tags[-1]
                logger.debug(f'The last output tag is {last_out_tag}')

                for idx, tag in enumerate(tags):
                    if not tag.is_the_same_as(last_out_tag):
                        continue

                    if not all(
                        map(
                            lambda t: t[0].is_the_same_as(t[1]),
                            zip(
                                tags[max(0, idx - (len(last_tags) - 1)) : idx],
                                last_tags[:-1],
                            ),
                        )
                    ):
                        continue

                    logger.debug(f'The last duplicated tag found at {idx} is {tag}')
                    return idx

                logger.debug('No duplicated tag found')
                return -1

            def update_delta_duplicated(last_duplicated_tag: FlvTag) -> None:
                nonlocal delta
                delta = last_tags[-1].timestamp - last_duplicated_tag.timestamp

            def update_delta_no_duplicated(first_data_tag: FlvTag) -> None:
                nonlocal delta
                delta = last_tags[-1].timestamp - first_data_tag.timestamp + 10

            def correct_ts(tag: FlvTag) -> FlvTag:
                if delta == 0:
                    return tag
                return tag.evolve(timestamp=tag.timestamp + delta)

            def make_join_point_tag(next_tag: FlvTag, seamless: bool) -> ScriptTag:
                assert next_tag.body is not None
                join_point = JoinPoint(
                    seamless=seamless,
                    timestamp=float(next_tag.timestamp),
                    crc32=cksum(next_tag.body),
                )
                logger.debug(f'join point: {join_point}; next tag: {next_tag}')
                script_data = ScriptData(
                    name='onJoinPoint', value=attr.asdict(join_point)
                )
                script_tag = create_script_tag(script_data)
                return script_tag

            def do_concat() -> None:
                logger.debug(
                    'Concatenating... gathered {} tags, total size: {}'.format(
                        len(gathered_tags), sum(t.tag_size for t in gathered_tags)
                    )
                )

                tags = list(
                    filter(
                        lambda tag: not is_metadata_tag(tag)
                        and not is_sequence_header(tag),
                        gathered_tags,
                    )
                )
                logger.debug(
                    '{} data tags, total size: {}'.format(
                        len(tags), sum(t.tag_size for t in tags)
                    )
                )

                if not tags:
                    return

                if (index := find_last_duplicated_tag(tags)) >= 0:
                    seamless = True
                    update_delta_duplicated(tags[index])
                    logger.debug(f'Updated delta: {delta}, seamless: {seamless}')
                    tags = tags[index + 1 :]
                else:
                    seamless = False
                    update_delta_no_duplicated(tags[0])
                    logger.debug(f'Updated delta: {delta}, seamless: {seamless}')

                if tags:
                    join_point_tag = make_join_point_tag(correct_ts(tags[0]), seamless)
                    observer.on_next(join_point_tag)

                for tag in tags:
                    tag = correct_ts(tag)
                    update_last_tags(tag)
                    observer.on_next(tag)

                gathered_tags.clear()

            def do_cancel() -> None:
                logger.debug(
                    'Cancelling... gathered {} tags, total size: {}'.format(
                        len(gathered_tags), sum(t.tag_size for t in gathered_tags)
                    )
                )
                assert last_flv_header is not None
                observer.on_next(last_flv_header)
                for tag in gathered_tags:
                    update_last_tags(tag)
                    observer.on_next(tag)
                gathered_tags.clear()

            def on_next(item: FLVStreamItem) -> None:
                nonlocal action
                nonlocal last_flv_header

                if isinstance(item, FlvHeader):
                    if last_flv_header is None:
                        logger.debug('No operation needed for the first stream')
                        last_flv_header = item
                        action = ACTION.NOOP
                        observer.on_next(item)
                    else:
                        logger.debug('Gathering tags for deduplication...')
                        last_flv_header = item
                        if action == ACTION.GATHER:
                            action = ACTION.CONCAT_AND_GATHER
                        else:
                            action = ACTION.GATHER
                    return

                tag = item

                while True:
                    if action == ACTION.NOOP:
                        update_last_tags(tag)
                        observer.on_next(tag)
                        return

                    if action == ACTION.CORRECT:
                        tag = correct_ts(tag)
                        update_last_tags(tag)
                        observer.on_next(tag)
                        return

                    if action in (ACTION.CONCAT, ACTION.CONCAT_AND_GATHER):
                        do_concat()
                        if action == ACTION.CONCAT_AND_GATHER:
                            action = ACTION.GATHER
                        else:
                            action = ACTION.CORRECT
                            return

                    if action == ACTION.GATHER:
                        gather_tags(tag)
                        if action == ACTION.CANCEL:
                            do_cancel()
                            action = ACTION.NOOP
                            return
                        if has_gathering_completed():
                            action = ACTION.CONCAT
                            continue

                    break

            def on_completed() -> None:
                if action == ACTION.GATHER:
                    do_concat()
                observer.on_completed()

            def on_error(e: Exception) -> None:
                if action == ACTION.GATHER:
                    do_concat()
                observer.on_error(e)

            return source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    return _concat


class JoinPointExtractor:
    def __init__(self) -> None:
        self._join_points: Subject[List[JoinPoint]] = Subject()

    @property
    def join_points(self) -> Observable[List[JoinPoint]]:
        return self._join_points

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._extract(source)

    def _extract(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            stream_index: int = -1
            subscription = SerialDisposable()
            join_points: List[JoinPoint] = []
            join_point_tag: Optional[ScriptTag] = None

            def push_join_points() -> None:
                self._join_points.on_next(join_points.copy())

            def on_next(item: FLVStreamItem) -> None:
                nonlocal stream_index
                nonlocal join_point_tag

                if isinstance(item, FlvHeader):
                    stream_index += 1
                    if stream_index > 0:
                        push_join_points()
                        join_points.clear()
                    join_point_tag = None
                    observer.on_next(item)
                    return

                if join_point_tag:
                    join_point = self._make_join_point(join_point_tag, item)
                    join_points.append(join_point)
                    join_point_tag = None

                if self._is_join_point_tag(item):
                    join_point_tag = item
                    return

                observer.on_next(item)

            def on_completed() -> None:
                push_join_points()
                observer.on_completed()

            def on_error(e: Exception) -> None:
                push_join_points()
                observer.on_error(e)

            def dispose() -> None:
                push_join_points()

            subscription.disposable = source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _is_join_point_tag(self, tag: FlvTag) -> TypeGuard[ScriptTag]:
        if is_script_tag(tag):
            script_data = parse_scriptdata(tag)
            return script_data['name'] == 'onJoinPoint'
        return False

    def _make_join_point(
        self, join_point_tag: ScriptTag, next_tag: FlvTag
    ) -> JoinPoint:
        script_data = parse_scriptdata(join_point_tag)
        join_point_data = cast(JoinPointData, script_data['value'])
        assert next_tag.body is not None, next_tag
        join_point = JoinPoint(
            seamless=join_point_data['seamless'],
            timestamp=next_tag.timestamp,
            crc32=join_point_data['crc32'],
        )
        logger.debug(f'Extracted join point: {join_point}; next tag: {next_tag}')
        return join_point



import logging


from .danmaku_client import DanmakuClient, DanmakuListener, DanmakuCommand
from .live import Live
from .typing import Danmaku
from .models import LiveStatus, RoomInfo
from ..event.event_emitter import EventListener, EventEmitter
from ..utils.mixins import SwitchableMixin


__all__ = 'LiveMonitor', 'LiveEventListener'


logger = logging.getLogger(__name__)


class LiveEventListener(EventListener):
    async def on_live_status_changed(
        self, current_status: LiveStatus, previous_status: LiveStatus
    ) -> None:
        ...

    async def on_live_began(self, live: Live) -> None:
        ...

    async def on_live_ended(self, live: Live) -> None:
        ...

    async def on_live_stream_available(self, live: Live) -> None:
        ...

    async def on_live_stream_reset(self, live: Live) -> None:
        ...

    async def on_room_changed(self, room_info: RoomInfo) -> None:
        ...


class LiveMonitor(
    EventEmitter[LiveEventListener], DanmakuListener, SwitchableMixin
):
    def __init__(self, danmaku_client: DanmakuClient, live: Live) -> None:
        super().__init__()
        self._danmaku_client = danmaku_client
        self._live = live

    def _init_status(self) -> None:
        self._previous_status = self._live.room_info.live_status
        if self._live.is_living():
            self._status_count = 2
        else:
            self._status_count = 0

    def _do_enable(self) -> None:
        self._init_status()
        self._danmaku_client.add_listener(self)
        logger.debug('Enabled live monitor')

    def _do_disable(self) -> None:
        self._danmaku_client.remove_listener(self)
        logger.debug('Disabled live monitor')

    async def on_client_reconnected(self) -> None:
        # check the live status after the client reconnected and simulate
        # events if necessary.
        # make sure the recorder works well continuously after interruptions
        # such as an operating system hibernation.
        logger.warning('The Danmaku Client Reconnected')

        await self._live.update_room_info()
        current_status = self._live.room_info.live_status

        if current_status == self._previous_status:
            if current_status == LiveStatus.LIVE:
                logger.debug('Simulating stream reset event')
                await self._handle_status_change(current_status)
        else:
            if current_status == LiveStatus.LIVE:
                logger.debug('Simulating live began event')
                await self._handle_status_change(current_status)
                logger.debug('Simulating live stream available event')
                await self._handle_status_change(current_status)
            else:
                logger.debug('Simulating live ended event')
                await self._handle_status_change(current_status)

    async def on_danmaku_received(self, danmu: Danmaku) -> None:
        danmu_cmd = danmu['cmd']

        if danmu_cmd == DanmakuCommand.LIVE.value:
            await self._handle_status_change(LiveStatus.LIVE)
        elif danmu_cmd == DanmakuCommand.PREPARING.value:
            if danmu.get('round', None) == 1:
                await self._handle_status_change(LiveStatus.ROUND)
            else:
                await self._handle_status_change(LiveStatus.PREPARING)
        elif danmu_cmd == DanmakuCommand.ROOM_CHANGE.value:
            await self._live.update_room_info()
            await self._emit('room_changed', self._live.room_info)

    async def _handle_status_change(self, current_status: LiveStatus) -> None:
        logger.debug('Live status changed from {} to {}'.format(
            self._previous_status.name, current_status.name
        ))

        await self._live.update_room_info()
        if (s := self._live.room_info.live_status) != current_status:
            logger.warning(
                'Updated live status {} is inconsistent with '
                'current live status {}'.format(s.name, current_status.name)
            )

        await self._emit(
            'live_status_changed', current_status, self._previous_status
        )

        if current_status != LiveStatus.LIVE:
            self._status_count = 0
            await self._emit('live_ended', self._live)
        else:
            self._status_count += 1

            if self._status_count == 1:
                assert self._previous_status != LiveStatus.LIVE
                await self._emit('live_began', self._live)
            elif self._status_count == 2:
                assert self._previous_status == LiveStatus.LIVE
                await self._emit('live_stream_available', self._live)
            elif self._status_count > 2:
                assert self._previous_status == LiveStatus.LIVE
                await self._emit('live_stream_reset', self._live)
            else:
                pass

        logger.debug('Number of sequential LIVE status: {}'.format(
            self._status_count
        ))

        self._previous_status = current_status



import logging
from abc import ABC, abstractmethod
from threading import Thread
from typing import Any, Iterator, List, Optional, Tuple, Union

import requests
import urllib3
from reactivex import abc
from reactivex.typing import StartableFactory, StartableTarget

from ..bili.live import Live
from ..bili.typing import QualityNumber, StreamFormat
from ..event.event_emitter import EventEmitter, EventListener
from ..flv import operators as flv_ops
from ..flv.metadata_dumper import MetadataDumper
from ..flv.operators import StreamProfile
from ..logging.room_id import aio_task_with_room_id
from ..utils.mixins import AsyncCooperationMixin, AsyncStoppableMixin
from . import operators as core_ops
from .metadata_provider import MetadataProvider
from .path_provider import PathProvider
from .stream_param_holder import StreamParamHolder

__all__ = ('StreamRecorderImpl',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class StreamRecorderEventListener(EventListener):
    async def on_video_file_created(self, path: str, record_start_time: int) -> None:
        ...

    async def on_video_file_completed(self, path: str) -> None:
        ...

    async def on_stream_recording_completed(self) -> None:
        ...


class StreamRecorderImpl(
    EventEmitter[StreamRecorderEventListener],
    AsyncCooperationMixin,
    AsyncStoppableMixin,
    ABC,
):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        stream_format: StreamFormat = 'flv',
        quality_number: QualityNumber = 10000,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__()

        self._live = live
        self._session = requests.Session()
        self._stream_param_holder = StreamParamHolder(
            stream_format=stream_format, quality_number=quality_number
        )
        self._stream_url_resolver = core_ops.StreamURLResolver(
            live, self._stream_param_holder
        )
        self._stream_fetcher = core_ops.StreamFetcher(
            live, self._session, read_timeout=read_timeout
        )
        self._stream_parser = core_ops.StreamParser(
            self._stream_param_holder,
            ignore_eof=stream_format != 'flv',
            ignore_value_error=stream_format != 'flv',
        )
        self._progress_bar = core_ops.ProgressBar(live)
        self._analyser = flv_ops.Analyser()
        self._metadata_provider = MetadataProvider(live, self)
        self._injector = flv_ops.Injector(self._metadata_provider)
        self._join_point_extractor = flv_ops.JoinPointExtractor()
        self._limiter = flv_ops.Limiter(filesize_limit, duration_limit)
        self._cutter = flv_ops.Cutter()
        self._path_provider = PathProvider(live, out_dir, path_template)
        self._dumper = flv_ops.Dumper(self._path_provider, buffer_size)
        self._rec_statistics = core_ops.SizedStatistics()

        self._prober: Union[flv_ops.Prober, core_ops.HLSProber]
        self._dl_statistics: Union[core_ops.StreamStatistics, core_ops.SizedStatistics]
        if stream_format == 'flv':
            self._prober = flv_ops.Prober()
            self._dl_statistics = core_ops.StreamStatistics()
        else:
            self._prober = core_ops.HLSProber()
            self._dl_statistics = core_ops.SizedStatistics()

        self._request_exception_handler = core_ops.RequestExceptionHandler()
        self._connection_error_handler = core_ops.ConnectionErrorHandler(
            live, disconnection_timeout=disconnection_timeout
        )
        self._exception_handler = core_ops.ExceptionHandler()
        self._metadata_dumper = MetadataDumper(
            self._dumper, self._analyser, self._join_point_extractor
        )
        self._metadata_dumper.enable()

        self._subscription: abc.DisposableBase
        self._completed: bool = False

        self._threads: List[Thread] = []
        self._files: List[str] = []
        self._stream_profile: StreamProfile = {}
        self._record_start_time: Optional[int] = None
        self._stream_available_time: Optional[int] = None
        self._hls_stream_available_time: Optional[int] = None

        def on_profile_updated(profile: StreamProfile) -> None:
            self._stream_profile = profile

        self._prober.profiles.subscribe(on_profile_updated)

        def on_file_opened(args: Tuple[str, int]) -> None:
            logger.info(f"Video file created: '{args[0]}'")
            self._files.append(args[0])
            self._record_start_time = args[1]
            self._emit_event('video_file_created', *args)

        def on_file_closed(path: str) -> None:
            logger.info(f"Video file completed: '{path}'")
            self._emit_event('video_file_completed', path)

        self._dumper.file_opened.subscribe(on_file_opened)
        self._dumper.file_closed.subscribe(on_file_closed)

    @property
    def stream_url(self) -> str:
        return self._stream_url_resolver.stream_url

    @property
    def stream_host(self) -> str:
        return self._stream_url_resolver.stream_host

    @property
    def record_start_time(self) -> Optional[int]:
        return self._record_start_time

    @property
    def stream_available_time(self) -> Optional[int]:
        return self._stream_available_time

    @stream_available_time.setter
    def stream_available_time(self, ts: Optional[int]) -> None:
        self._stream_available_time = ts

    @property
    def hls_stream_available_time(self) -> Optional[int]:
        return self._hls_stream_available_time

    @hls_stream_available_time.setter
    def hls_stream_available_time(self, ts: Optional[int]) -> None:
        self._hls_stream_available_time = ts

    @property
    def dl_total(self) -> int:
        return self._dl_statistics.count

    @property
    def dl_rate(self) -> float:
        return self._dl_statistics.rate

    @property
    def rec_elapsed(self) -> float:
        return self._rec_statistics.elapsed

    @property
    def rec_total(self) -> int:
        return self._rec_statistics.count

    @property
    def rec_rate(self) -> float:
        return self._rec_statistics.rate

    @property
    def out_dir(self) -> str:
        return self._path_provider.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._path_provider.out_dir = value

    @property
    def path_template(self) -> str:
        return self._path_provider.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._path_provider.path_template = value

    @property
    def stream_format(self) -> StreamFormat:
        return self._stream_param_holder.stream_format

    @property
    def quality_number(self) -> QualityNumber:
        return self._stream_param_holder.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._stream_param_holder.quality_number = value

    @property
    def real_quality_number(self) -> Optional[QualityNumber]:
        if self.stopped:
            return None
        return self._stream_param_holder.real_quality_number

    @property
    def filesize_limit(self) -> int:
        return self._limiter.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._limiter.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._limiter.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._limiter.duration_limit = value

    @property
    def read_timeout(self) -> int:
        return self._stream_fetcher.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._stream_fetcher.read_timeout = value

    @property
    def disconnection_timeout(self) -> int:
        return self._connection_error_handler.disconnection_timeout

    @disconnection_timeout.setter
    def disconnection_timeout(self, value: int) -> None:
        self._connection_error_handler.disconnection_timeout = value

    @property
    def buffer_size(self) -> int:
        return self._dumper.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._dumper.buffer_size = value

    @property
    def recording_path(self) -> Optional[str]:
        return self._dumper.path

    @property
    def metadata(self) -> Optional[flv_ops.MetaData]:
        try:
            return self._analyser.make_metadata()
        except Exception:
            return None

    @property
    def stream_profile(self) -> StreamProfile:
        return self._stream_profile

    def has_file(self) -> bool:
        return bool(self._files)

    def get_files(self) -> Iterator[str]:
        yield from iter(self._files)

    def clear_files(self) -> None:
        self._files.clear()

    def can_cut_stream(self) -> bool:
        return self._cutter.can_cut_stream()

    def cut_stream(self) -> bool:
        return self._cutter.cut_stream()

    def update_progress_bar_info(self) -> None:
        self._progress_bar.update_bar_info()

    def _reset(self) -> None:
        self._files.clear()
        self._stream_profile = {}
        self._record_start_time = None
        self._completed = False

    async def _do_start(self) -> None:
        logger.debug('Starting stream recorder...')
        self._reset()
        self._run()
        logger.debug('Started stream recorder')

    async def _do_stop(self) -> None:
        logger.debug('Stopping stream recorder...')
        self._stream_param_holder.cancel()
        thread = self._thread_factory('StreamRecorderDisposer')(self._dispose)
        thread.start()
        for thread in self._threads:
            await self._loop.run_in_executor(None, thread.join, 30)
        self._threads.clear()
        logger.debug('Stopped stream recorder')

    @abstractmethod
    def _run(self) -> None:
        raise NotImplementedError()

    def _thread_factory(self, name: str) -> StartableFactory:
        def factory(target: StartableTarget) -> Thread:
            thread = Thread(
                target=target, daemon=True, name=f'{name}::{self._live.room_id}'
            )
            self._threads.append(thread)
            return thread

        return factory

    def _dispose(self) -> None:
        self._subscription.dispose()
        self._on_completed()

    def _on_completed(self) -> None:
        if self._completed:
            return
        self._completed = True

        self._dl_statistics.freeze()
        self._rec_statistics.freeze()
        self._emit_event('stream_recording_completed')

    def _emit_event(self, name: str, *args: Any, **kwds: Any) -> None:
        self._run_coroutine(self._emit(name, *args, **kwds))

    @aio_task_with_room_id
    async def _emit(self, *args: Any, **kwds: Any) -> None:  # type: ignore
        await super()._emit(*args, **kwds)



from __future__ import annotations

import logging
import os
import re
from typing import ClassVar, Collection, Final, List, Optional, TypeVar

import toml
from pydantic import BaseModel as PydanticBaseModel
from pydantic import BaseSettings, Field, PrivateAttr, validator
from pydantic.networks import EmailStr, HttpUrl
from typing_extensions import Annotated

from ..bili.typing import QualityNumber, StreamFormat
from ..core.cover_downloader import CoverSaveStrategy
from ..logging.typing import LOG_LEVEL
from ..postprocess import DeleteStrategy
from ..utils.string import camel_case
from .typing import (
    EmailMessageType,
    PushdeerMessageType,
    PushplusMessageType,
    ServerchanMessageType,
    TelegramMessageType,
)

logger = logging.getLogger(__name__)


__all__ = (
    'DEFAULT_SETTINGS_FILE',
    'EnvSettings',
    'Settings',
    'SettingsIn',
    'SettingsOut',
    'HeaderOptions',
    'HeaderSettings',
    'DanmakuOptions',
    'DanmakuSettings',
    'RecorderOptions',
    'RecorderSettings',
    'PostprocessingSettings',
    'PostprocessingOptions',
    'TaskOptions',
    'TaskSettings',
    'OutputSettings',
    'LoggingSettings',
    'SpaceSettings',
    'EmailSettings',
    'ServerchanSettings',
    'PushdeerSettings',
    'PushplusSettings',
    'TelegramSettings',
    'NotifierSettings',
    'NotificationSettings',
    'EmailMessageTemplateSettings',
    'ServerchanMessageTemplateSettings',
    'PushdeerMessageTemplateSettings',
    'PushplusMessageTemplateSettings',
    'TelegramMessageTemplateSettings',
    'EmailNotificationSettings',
    'ServerchanNotificationSettings',
    'PushdeerNotificationSettings',
    'PushplusNotificationSettings',
    'TelegramNotificationSettings',
    'WebHookSettings',
)


DEFAULT_OUT_DIR: Final[str] = os.environ.get('DEFAULT_OUT_DIR', '.')
DEFAULT_LOG_DIR: Final[str] = os.environ.get('DEFAULT_LOG_DIR', '~/.blrec/logs/')
DEFAULT_SETTINGS_FILE: Final[str] = os.environ.get(
    'DEFAULT_SETTINGS_FILE', '~/.blrec/settings.toml'
)


class EnvSettings(BaseSettings):
    settings_file: Annotated[str, Field(env='config')] = DEFAULT_SETTINGS_FILE
    out_dir: Optional[str] = None
    log_dir: Optional[str] = None
    api_key: Annotated[
        Optional[str], Field(min_length=8, max_length=80, regex=r'[a-zA-Z\d\-]{8,80}'),
    ] = None

    class Config:
        anystr_strip_whitespace = True


_V = TypeVar('_V')


class BaseModel(PydanticBaseModel):
    class Config:
        validate_assignment = True
        anystr_strip_whitespace = True
        allow_population_by_field_name = True

        @classmethod
        def alias_generator(cls, string: str) -> str:
            return camel_case(string)

    @staticmethod
    def _validate_with_collection(value: _V, allowed_values: Collection[_V]) -> None:
        if value not in allowed_values:
            raise ValueError(
                f'the value {value} does not be allowed, '
                f'must be one of {", ".join(map(str, allowed_values))}'
            )


class HeaderOptions(BaseModel):
    user_agent: Optional[str]
    cookie: Optional[str]


class HeaderSettings(HeaderOptions):
    user_agent: str = (
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
        'AppleWebKit/537.36 (KHTML, like Gecko) '
        'Chrome/89.0.4389.114 Safari/537.36'
    )
    cookie: str = ''


class DanmakuOptions(BaseModel):
    danmu_uname: Optional[bool]
    record_gift_send: Optional[bool]
    record_free_gifts: Optional[bool]
    record_guard_buy: Optional[bool]
    record_super_chat: Optional[bool]
    save_raw_danmaku: Optional[bool]


class DanmakuSettings(DanmakuOptions):
    danmu_uname: bool = False
    record_gift_send: bool = True
    record_free_gifts: bool = True
    record_guard_buy: bool = True
    record_super_chat: bool = True
    save_raw_danmaku: bool = False


class RecorderOptions(BaseModel):
    stream_format: Optional[StreamFormat]
    quality_number: Optional[QualityNumber]
    fmp4_stream_timeout: Optional[int]
    read_timeout: Optional[int]  # seconds
    disconnection_timeout: Optional[int]  # seconds
    buffer_size: Annotated[  # bytes
        Optional[int], Field(ge=4096, le=1024**2 * 512, multiple_of=2)
    ]
    save_cover: Optional[bool]
    cover_save_strategy: Optional[CoverSaveStrategy]

    @validator('fmp4_stream_timeout')
    def _validate_fmp4_stream_timeout(cls, v: Optional[int]) -> Optional[int]:
        if v is not None:
            allowed_values = frozenset((3, 5, 10, 30, 60, 180, 300, 600))
            cls._validate_with_collection(v, allowed_values)
        return v

    @validator('read_timeout')
    def _validate_read_timeout(cls, value: Optional[int]) -> Optional[int]:
        if value is not None:
            allowed_values = frozenset((3, 5, 10, 30, 60, 180, 300, 600))
            cls._validate_with_collection(value, allowed_values)
        return value

    @validator('disconnection_timeout')
    def _validate_disconnection_timeout(cls, value: Optional[int]) -> Optional[int]:
        if value is not None:
            allowed_values = frozenset(60 * i for i in (3, 5, 10, 15, 20, 30))
            cls._validate_with_collection(value, allowed_values)
        return value


class RecorderSettings(RecorderOptions):
    stream_format: StreamFormat = 'flv'
    quality_number: QualityNumber = 20000  # 4K, the highest quality.
    fmp4_stream_timeout: int = 10
    read_timeout: int = 3
    disconnection_timeout: int = 600
    buffer_size: Annotated[
        int, Field(ge=4096, le=1024**2 * 512, multiple_of=2)
    ] = 8192
    save_cover: bool = False
    cover_save_strategy: CoverSaveStrategy = CoverSaveStrategy.DEFAULT


class PostprocessingOptions(BaseModel):
    remux_to_mp4: Optional[bool]
    inject_extra_metadata: Optional[bool]
    delete_source: Optional[DeleteStrategy]


class PostprocessingSettings(PostprocessingOptions):
    remux_to_mp4: bool = False
    inject_extra_metadata: bool = True
    delete_source: DeleteStrategy = DeleteStrategy.AUTO


class OutputOptions(BaseModel):
    path_template: Optional[str]
    filesize_limit: Optional[int]  # file size in bytes
    duration_limit: Optional[int]  # duration in seconds

    @validator('path_template')
    def _validate_path_template(cls, value: Optional[str]) -> Optional[str]:
        if value is not None:
            pattern = r'''^
                (?:
                    [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                    \{
                    (?:
                        roomid|uname|title|area|parent_area|
                        year|month|day|hour|minute|second
                    )
                    \}
                    [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                )+?
                (?:
                    /
                    (?:
                        [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                        \{
                        (?:
                            roomid|uname|title|area|parent_area|
                            year|month|day|hour|minute|second
                        )
                        \}
                        [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                    )+?
                )*
            $'''
            if not re.fullmatch(pattern, value, re.VERBOSE):
                raise ValueError(f"invalid path template: '{value}'")
        return value

    @validator('filesize_limit')
    def _validate_filesize_limit(cls, value: Optional[int]) -> Optional[int]:
        # allowed 1 ~ 20 GB, 0 indicates not limit.
        if value is not None:
            allowed_values = frozenset(1024**3 * i for i in range(0, 21))
            cls._validate_with_collection(value, allowed_values)
        return value

    @validator('duration_limit')
    def _validate_duration_limit(cls, value: Optional[int]) -> Optional[int]:
        # allowed 1 ~ 24 hours, 0 indicates not limit.
        if value is not None:
            allowed_values = frozenset(3600 * i for i in range(0, 25))
            cls._validate_with_collection(value, allowed_values)
        return value


def out_dir_factory() -> str:
    path = os.path.normpath(os.path.expanduser(DEFAULT_OUT_DIR))
    os.makedirs(path, exist_ok=True)
    return path


class OutputSettings(OutputOptions):
    out_dir: Annotated[str, Field(default_factory=out_dir_factory)]
    path_template: str = (
        '{roomid} - {uname}/'
        'blive_{roomid}_{year}-{month}-{day}-{hour}{minute}{second}'
    )
    filesize_limit: int = 0  # no limit by default
    duration_limit: int = 0  # no limit by default

    @validator('out_dir')
    def _validate_dir(cls, path: str) -> str:
        if not os.path.isdir(os.path.expanduser(path)):
            raise ValueError(f"'{path}' not a directory")
        return path


class TaskOptions(BaseModel):
    output: OutputOptions = OutputOptions()
    header: HeaderOptions = HeaderOptions()
    danmaku: DanmakuOptions = DanmakuOptions()
    recorder: RecorderOptions = RecorderOptions()
    postprocessing: PostprocessingOptions = PostprocessingOptions()

    @classmethod
    def from_settings(cls, settings: TaskSettings) -> TaskOptions:
        return cls(
            **settings.dict(
                include={'output', 'header', 'danmaku', 'recorder', 'postprocessing'}
            )
        )


class TaskSettings(TaskOptions):
    # must use the real room id rather than the short room id!
    room_id: Annotated[int, Field(ge=1, le=99999999)]
    enable_monitor: bool = True
    enable_recorder: bool = True


def log_dir_factory() -> str:
    path = os.path.normpath(os.path.expanduser(DEFAULT_LOG_DIR))
    os.makedirs(path, exist_ok=True)
    return path


class LoggingSettings(BaseModel):
    log_dir: Annotated[str, Field(default_factory=log_dir_factory)]
    console_log_level: LOG_LEVEL = 'INFO'
    max_bytes: Annotated[
        int, Field(ge=1024**2, le=1024**2 * 10, multiple_of=1024**2)
    ] = (
        1024**2 * 10
    )  # allowed 1 ~ 10 MB
    backup_count: Annotated[int, Field(ge=1, le=30)] = 30

    @validator('log_dir')
    def _validate_dir(cls, path: str) -> str:
        if not os.path.isdir(os.path.expanduser(path)):
            raise ValueError(f"'{path}' not a directory")
        return path


class SpaceSettings(BaseModel):
    check_interval: int = 60  # 1 minutes
    space_threshold: int = 1024**3  # 1 GB
    recycle_records: bool = False

    @validator('check_interval')
    def _validate_interval(cls, value: int) -> int:
        allowed_values = frozenset((10, 30, *(60 * i for i in (1, 3, 5, 10))))
        cls._validate_with_collection(value, allowed_values)
        return value

    @validator('space_threshold')
    def _validate_threshold(cls, value: int) -> int:
        allowed_values = frozenset(1024**3 * i for i in (1, 3, 5, 10, 20))
        cls._validate_with_collection(value, allowed_values)
        return value


class EmailSettings(BaseModel):
    src_addr: Annotated[str, EmailStr] = ''
    dst_addr: Annotated[str, EmailStr] = ''
    auth_code: str = ''
    smtp_host: str = 'smtp.163.com'
    smtp_port: int = 465


class ServerchanSettings(BaseModel):
    sendkey: str = ''

    @validator('sendkey')
    def _validate_sendkey(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'[a-zA-Z\d]+', value):
            raise ValueError('sendkey is invalid')
        return value


class PushdeerSettings(BaseModel):
    server: str = ''
    pushkey: str = ''

    @validator('server')
    def _validate_server(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'https?://.+', value):
            raise ValueError('server is invalid')
        return value

    @validator('pushkey')
    def _validate_pushkey(cls, value: str) -> str:
        if value != '' and not re.fullmatch(
            r'PDU\d+T[a-zA-Z\d]{32}(,PDU\d+T[a-zA-Z\d]{32}){0,99}', value
        ):
            raise ValueError('pushkey is invalid')
        return value


class PushplusSettings(BaseModel):
    token: str = ''
    topic: str = ''

    @validator('token')
    def _validate_token(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'[a-z\d]{32}', value):
            raise ValueError('token is invalid')
        return value


class TelegramSettings(BaseModel):
    token: str = ''
    chatid: str = ''

    @validator('token')
    def _validate_token(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'[0-9]{8,10}:[a-zA-Z0-9_-]{35}', value):
            raise ValueError('token is invalid')
        return value

    @validator('chatid')
    def _validate_chatid(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'(-|[0-9]){0,}', value):
            raise ValueError('chatid is invalid')
        return value


class NotifierSettings(BaseModel):
    enabled: bool = False


class NotificationSettings(BaseModel):
    notify_began: bool = True
    notify_ended: bool = True
    notify_error: bool = True
    notify_space: bool = True


class MessageTemplateSettings(BaseModel):
    began_message_type: str
    began_message_title: str
    began_message_content: str
    ended_message_type: str
    ended_message_title: str
    ended_message_content: str
    space_message_type: str
    space_message_title: str
    space_message_content: str
    error_message_type: str
    error_message_title: str
    error_message_content: str


class EmailMessageTemplateSettings(MessageTemplateSettings):
    began_message_type: EmailMessageType = 'html'
    began_message_title: str = ''
    began_message_content: str = ''
    ended_message_type: EmailMessageType = 'html'
    ended_message_title: str = ''
    ended_message_content: str = ''
    space_message_type: EmailMessageType = 'html'
    space_message_title: str = ''
    space_message_content: str = ''
    error_message_type: EmailMessageType = 'html'
    error_message_title: str = ''
    error_message_content: str = ''


class ServerchanMessageTemplateSettings(MessageTemplateSettings):
    began_message_type: ServerchanMessageType = 'markdown'
    began_message_title: str = ''
    began_message_content: str = ''
    ended_message_type: ServerchanMessageType = 'markdown'
    ended_message_title: str = ''
    ended_message_content: str = ''
    space_message_type: ServerchanMessageType = 'markdown'
    space_message_title: str = ''
    space_message_content: str = ''
    error_message_type: ServerchanMessageType = 'markdown'
    error_message_title: str = ''
    error_message_content: str = ''


class PushdeerMessageTemplateSettings(MessageTemplateSettings):
    began_message_type: PushdeerMessageType = 'markdown'
    began_message_title: str = ''
    began_message_content: str = ''
    ended_message_type: PushdeerMessageType = 'markdown'
    ended_message_title: str = ''
    ended_message_content: str = ''
    space_message_type: PushdeerMessageType = 'markdown'
    space_message_title: str = ''
    space_message_content: str = ''
    error_message_type: PushdeerMessageType = 'markdown'
    error_message_title: str = ''
    error_message_content: str = ''


class PushplusMessageTemplateSettings(MessageTemplateSettings):
    began_message_type: PushplusMessageType = 'markdown'
    began_message_title: str = ''
    began_message_content: str = ''
    ended_message_type: PushplusMessageType = 'markdown'
    ended_message_title: str = ''
    ended_message_content: str = ''
    space_message_type: PushplusMessageType = 'markdown'
    space_message_title: str = ''
    space_message_content: str = ''
    error_message_type: PushplusMessageType = 'markdown'
    error_message_title: str = ''
    error_message_content: str = ''


class TelegramMessageTemplateSettings(MessageTemplateSettings):
    began_message_type: TelegramMessageType = 'html'
    began_message_title: str = ''
    began_message_content: str = ''
    ended_message_type: TelegramMessageType = 'html'
    ended_message_title: str = ''
    ended_message_content: str = ''
    space_message_type: TelegramMessageType = 'html'
    space_message_title: str = ''
    space_message_content: str = ''
    error_message_type: TelegramMessageType = 'html'
    error_message_title: str = ''
    error_message_content: str = ''


class EmailNotificationSettings(
    EmailSettings, NotifierSettings, NotificationSettings, EmailMessageTemplateSettings
):
    pass


class ServerchanNotificationSettings(
    ServerchanSettings,
    NotifierSettings,
    NotificationSettings,
    ServerchanMessageTemplateSettings,
):
    pass


class PushdeerNotificationSettings(
    PushdeerSettings,
    NotifierSettings,
    NotificationSettings,
    PushplusMessageTemplateSettings,
):
    pass


class PushplusNotificationSettings(
    PushplusSettings,
    NotifierSettings,
    NotificationSettings,
    PushplusMessageTemplateSettings,
):
    pass


class TelegramNotificationSettings(
    TelegramSettings,
    NotifierSettings,
    NotificationSettings,
    TelegramMessageTemplateSettings,
):
    pass


class WebHookEventSettings(BaseModel):
    live_began: bool = True
    live_ended: bool = True
    room_change: bool = True
    recording_started: bool = True
    recording_finished: bool = True
    recording_cancelled: bool = True
    video_file_created: bool = True
    video_file_completed: bool = True
    danmaku_file_created: bool = True
    danmaku_file_completed: bool = True
    raw_danmaku_file_created: bool = True
    raw_danmaku_file_completed: bool = True
    video_postprocessing_completed: bool = True
    space_no_enough: bool = True
    error_occurred: bool = True


class WebHookSettings(WebHookEventSettings):
    url: Annotated[str, HttpUrl]


class Settings(BaseModel):
    _MAX_TASKS: ClassVar[int] = 100
    _MAX_WEBHOOKS: ClassVar[int] = 50

    _path: str = PrivateAttr()
    version: str = '1.0'

    tasks: Annotated[List[TaskSettings], Field(max_items=100)] = []
    output: OutputSettings = OutputSettings()  # type: ignore
    logging: LoggingSettings = LoggingSettings()  # type: ignore
    header: HeaderSettings = HeaderSettings()
    danmaku: DanmakuSettings = DanmakuSettings()
    recorder: RecorderSettings = RecorderSettings()
    postprocessing: PostprocessingSettings = PostprocessingSettings()
    space: SpaceSettings = SpaceSettings()
    email_notification: EmailNotificationSettings = EmailNotificationSettings()
    serverchan_notification: ServerchanNotificationSettings = (
        ServerchanNotificationSettings()
    )
    pushdeer_notification: PushdeerNotificationSettings = PushdeerNotificationSettings()
    pushplus_notification: PushplusNotificationSettings = PushplusNotificationSettings()
    telegram_notification: TelegramNotificationSettings = TelegramNotificationSettings()
    webhooks: Annotated[List[WebHookSettings], Field(max_items=50)] = []

    @classmethod
    def load(cls, path: str) -> Settings:
        settings = cls.parse_obj(toml.load(path))
        settings._path = path
        return settings

    def update_from_env_settings(self, env_settings: EnvSettings) -> None:
        if (out_dir := env_settings.out_dir) is not None:
            self.output.out_dir = out_dir
        if (log_dir := env_settings.log_dir) is not None:
            self.logging.log_dir = log_dir

    def dump(self) -> None:
        assert self._path
        with open(self._path, 'wt', encoding='utf8') as file:
            toml.dump(self.dict(exclude_none=True), file)

    @validator('tasks')
    def _validate_tasks(cls, tasks: List[TaskSettings]) -> List[TaskSettings]:
        if len(tasks) > cls._MAX_TASKS:
            raise ValueError(f'Out of max tasks limits: {cls._MAX_TASKS}')
        return tasks

    @validator('webhooks')
    def _validate_webhooks(
        cls, webhooks: List[WebHookSettings]
    ) -> List[WebHookSettings]:
        if len(webhooks) >= cls._MAX_WEBHOOKS:
            raise ValueError(f'Out of max webhooks limits: {cls._MAX_WEBHOOKS}')
        return webhooks


class SettingsIn(BaseModel):
    output: Optional[OutputSettings] = None
    logging: Optional[LoggingSettings] = None
    header: Optional[HeaderSettings] = None
    danmaku: Optional[DanmakuSettings] = None
    recorder: Optional[RecorderSettings] = None
    postprocessing: Optional[PostprocessingSettings] = None
    space: Optional[SpaceSettings] = None
    email_notification: Optional[EmailNotificationSettings] = None
    serverchan_notification: Optional[ServerchanNotificationSettings] = None
    pushdeer_notification: Optional[PushdeerNotificationSettings] = None
    pushplus_notification: Optional[PushplusNotificationSettings] = None
    telegram_notification: Optional[TelegramNotificationSettings] = None
    webhooks: Optional[List[WebHookSettings]] = None


class SettingsOut(SettingsIn):
    version: Optional[str] = None
    tasks: Optional[List[TaskSettings]] = None



from .exceptions import (
    ExistsError,
    NotFoundError,
    ForbiddenError,
)
from .exception_center import ExceptionCenter
from .exception_handler import ExceptionHandler
from .exception_submiter import (
    ExceptionSubmitter,
    submit_exception,
    exception_callback,
)
from .helpers import format_exception


__all__ = (
    'ExistsError',
    'NotFoundError',
    'ForbiddenError',

    'ExceptionCenter',
    'ExceptionSubmitter',
    'submit_exception',
    'exception_callback',

    'ExceptionHandler',

    'format_exception',
)



"""
Analyse the FLV stream to make MetaData
ref: https://github.com/ioppermann/yamdi/blob/master/yamdi.c
"""
from __future__ import annotations

import logging
from typing import List, Optional, TypedDict

import attr
from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ..common import Resolution, is_audio_tag, is_script_tag, is_video_tag
from ..models import (
    BACK_POINTER_SIZE,
    AudioTag,
    FlvHeader,
    FlvTag,
    ScriptTag,
    SoundType,
    VideoTag,
)
from .typing import FLVStream, FLVStreamItem

__all__ = 'Analyser', 'MetaData', 'KeyFrames'

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class KeyFrames:
    times: List[float]
    filepositions: List[float]


class KeyFramesDict(TypedDict):
    times: List[float]
    filepositions: List[float]


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class MetaData:
    hasAudio: bool
    hasVideo: bool
    hasMetadata: bool
    hasKeyframes: bool
    canSeekToEnd: bool
    duration: float
    datasize: float
    filesize: float

    audiosize: Optional[float] = None
    audiocodecid: Optional[float] = None
    audiodatarate: Optional[float] = None
    audiosamplerate: Optional[float] = None
    audiosamplesize: Optional[float] = None
    stereo: Optional[bool] = None

    videosize: float
    framerate: float
    videocodecid: float
    videodatarate: float
    width: float
    height: float

    lasttimestamp: float
    lastkeyframelocation: float
    lastkeyframetimestamp: float
    keyframes: KeyFrames


class MetaDataDict:
    hasAudio: bool
    hasVideo: bool
    hasMetadata: bool
    hasKeyframes: bool
    canSeekToEnd: bool
    duration: float
    datasize: float
    filesize: float

    audiosize: Optional[float] = None
    audiocodecid: Optional[float] = None
    audiodatarate: Optional[float] = None
    audiosamplerate: Optional[float] = None
    audiosamplesize: Optional[float] = None
    stereo: Optional[bool] = None

    videosize: float
    framerate: float
    videocodecid: float
    videodatarate: float
    width: float
    height: float

    lasttimestamp: float
    lastkeyframelocation: float
    lastkeyframetimestamp: float
    keyframes: KeyFramesDict


class Analyser:
    def __init__(self) -> None:
        self._metadatas: Subject[Optional[MetaData]] = Subject()
        self._reset()

    def _reset(self) -> None:
        self._num_of_tags = 0
        self._num_of_audio_tags = 0
        self._num_of_video_tags = 0
        self._size_of_tags = 0
        self._size_of_audio_tags = 0
        self._size_of_video_tags = 0
        self._size_of_data = 0
        self._size_of_audio_data = 0
        self._size_of_video_data = 0
        self._last_timestamp = 0
        self._last_timestamp_of_audio = 0
        self._last_timestamp_of_video = 0
        self._keyframe_timestamps: List[int] = []
        self._keyframe_filepositions: List[int] = []
        self._resolution: Optional[Resolution] = None

        self._has_audio = False
        self._has_video = False
        self._audio_analysed = False
        self._video_analysed = False

    @property
    def metadatas(self) -> Observable[MetaData]:
        return self._metadatas

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._analyse(source)

    def get_real_resolution(self) -> Optional[Resolution]:
        return self._resolution

    def calc_frame_rate(self) -> float:
        try:
            return self._num_of_video_tags / self._last_timestamp_of_video * 1000
        except ZeroDivisionError:
            return 0.0

    def calc_audio_data_rate(self) -> float:
        try:
            return self._size_of_audio_data * 8 / self._last_timestamp_of_audio
        except ZeroDivisionError:
            return 0.0

    def calc_video_data_rate(self) -> float:
        try:
            return self._size_of_video_data * 8 / self._last_timestamp_of_video
        except ZeroDivisionError:
            return 0.0

    def calc_data_size(self) -> int:
        return (
            self._size_of_audio_tags
            + self._num_of_audio_tags * BACK_POINTER_SIZE
            + self._size_of_video_tags
            + self._num_of_video_tags * BACK_POINTER_SIZE
        )

    def calc_file_size(self) -> int:
        return (
            self._size_of_flv_header
            + BACK_POINTER_SIZE
            + self._size_of_tags
            + self._num_of_tags * BACK_POINTER_SIZE
        )

    def make_keyframes(self) -> KeyFrames:
        return KeyFrames(
            times=list(map(lambda t: t / 1000, self._keyframe_timestamps)),
            filepositions=list(map(float, self._keyframe_filepositions)),
        )

    def make_metadata(self) -> MetaData:
        assert self._has_audio == self._audio_analysed, (
            f'has_audio: {self._has_audio}, audio_analysed: {self._audio_analysed}',
        )
        assert self._has_video and self._video_analysed, (
            f'has_video: {self._has_video}, video_analysed: {self._video_analysed}',
        )
        assert self._resolution is not None, 'no resolution'

        if not self._has_audio:
            audiosize = None
            audiocodecid = None
            audiodatarate = None
            audiosamplerate = None
            audiosamplesize = None
            stereo = None
        else:
            audiosize = float(self._size_of_audio_tags)
            audiocodecid = float(self._audio_codec_id)
            audiodatarate = self.calc_audio_data_rate()
            audiosamplerate = float(self._audio_sample_rate)
            audiosamplesize = float(self._audio_sample_size)
            stereo = self._stereo

        keyframes = self.make_keyframes()

        return MetaData(
            hasAudio=self._has_audio,
            hasVideo=self._has_video,
            hasMetadata=True,
            hasKeyframes=len(self._keyframe_timestamps) != 0,
            canSeekToEnd=(
                self._last_timestamp_of_video == self._keyframe_timestamps[-1]
            ),
            duration=self._last_timestamp / 1000,
            datasize=float(self.calc_data_size()),
            filesize=float(self.calc_file_size()),
            audiosize=audiosize,
            audiocodecid=audiocodecid,
            audiodatarate=audiodatarate,
            audiosamplerate=audiosamplerate,
            audiosamplesize=audiosamplesize,
            stereo=stereo,
            videosize=float(self._size_of_video_tags),
            framerate=self.calc_frame_rate(),
            videocodecid=float(self._video_codec_id),
            videodatarate=self.calc_video_data_rate(),
            width=float(self._resolution.width),
            height=float(self._resolution.height),
            lasttimestamp=self._last_timestamp / 1000,
            lastkeyframelocation=keyframes.filepositions[-1],
            lastkeyframetimestamp=keyframes.times[-1],
            keyframes=keyframes,
        )

    def _analyse(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            stream_index: int = -1
            subscription = SerialDisposable()

            def push_metadata() -> None:
                try:
                    metadata = self.make_metadata()
                except Exception as e:
                    logger.warning(f'Failed to make metadata: {repr(e)}')
                    self._metadatas.on_next(None)
                else:
                    self._metadatas.on_next(metadata)

            def on_next(item: FLVStreamItem) -> None:
                nonlocal stream_index
                if isinstance(item, FlvHeader):
                    stream_index += 1
                    if stream_index > 0:
                        push_metadata()
                    self._reset()
                    self._analyse_flv_header(item)
                else:
                    self._analyse_tag(item)
                observer.on_next(item)

            def on_completed() -> None:
                push_metadata()
                observer.on_completed()

            def on_error(e: Exception) -> None:
                push_metadata()
                observer.on_error(e)

            def dispose() -> None:
                push_metadata()

            subscription.disposable = source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _analyse_flv_header(self, header: FlvHeader) -> None:
        self._size_of_flv_header = header.size

    def _analyse_tag(self, tag: FlvTag) -> None:
        if is_audio_tag(tag):
            self._analyse_audio_tag(tag)
        elif is_video_tag(tag):
            self._analyse_video_tag(tag)
        elif is_script_tag(tag):
            self._analyse_script_tag(tag)
        else:
            logger.warning(f'Invalid tag type: {tag}')

        self._num_of_tags += 1
        self._size_of_tags += tag.tag_size
        self._size_of_data += tag.data_size
        self._last_timestamp = tag.timestamp

    def _analyse_audio_tag(self, tag: AudioTag) -> None:
        if not self._audio_analysed:
            self._has_audio = True
            self._audio_analysed = True
            self._audio_codec_id = tag.sound_format.value
            self._audio_sample_rate = tag.sound_rate.value
            self._audio_sample_size = tag.sound_size.value
            self._stereo = tag.sound_type == SoundType.STEREO
            logger.debug(f'Audio analysed: {tag}')

        self._num_of_audio_tags += 1
        self._size_of_audio_tags += tag.tag_size
        self._size_of_audio_data += tag.data_size
        self._last_timestamp_of_audio = tag.timestamp

    def _analyse_video_tag(self, tag: VideoTag) -> None:
        if tag.is_keyframe():
            self._keyframe_timestamps.append(tag.timestamp)
            self._keyframe_filepositions.append(self.calc_file_size())
            if tag.is_avc_header():
                self._resolution = Resolution.from_aac_sequence_header(tag)
                logger.debug(f'Resolution: {self._resolution}')
        else:
            pass

        if not self._video_analysed:
            self._has_video = True
            self._video_analysed = True
            self._video_codec_id = tag.codec_id.value
            logger.debug(f'Video analysed: {tag}')

        self._num_of_video_tags += 1
        self._size_of_video_tags += tag.tag_size
        self._size_of_video_data += tag.data_size
        self._last_timestamp_of_video = tag.timestamp

    def _analyse_script_tag(self, tag: ScriptTag) -> None:
        pass



import logging
from typing import Iterable


from .io import DanmakuReader, DanmakuWriter
from .common import copy_damus


logger = logging.getLogger(__name__)


__all__ = 'DanmakuConcatenator'


class DanmakuConcatenator:
    def __init__(
        self, in_paths: Iterable[str], deltas: Iterable[int], out_path: str
    ) -> None:
        self._in_paths = list(in_paths)
        self._deltas = list(deltas)
        self._out_path = out_path
        assert len(self._in_paths) == len(self._deltas)

    async def concat(self) -> None:
        async with DanmakuWriter(self._out_path) as writer:
            async with DanmakuReader(self._in_paths[0]) as reader:
                metadata = await reader.read_metadata()
                await writer.write_metadata(metadata)
                await copy_damus(reader, writer, delta=self._deltas[0])

            for path, delta in zip(self._in_paths[1:], self._deltas[1:]):
                async with DanmakuReader(path) as reader:
                    await copy_damus(reader, writer, delta=delta)



from __future__ import annotations

import logging
from typing import Any, Callable, Dict, Optional, cast

from reactivex import Observable, abc

from ..common import (
    create_metadata_tag,
    enrich_metadata,
    is_metadata_tag,
    parse_metadata,
    update_metadata,
)
from ..models import FlvHeader, ScriptTag
from .analyse import KeyFramesDict
from .typing import FLVStream, FLVStreamItem

__all__ = ('Injector',)


logger = logging.getLogger(__name__)


class Injector:
    def __init__(
        self, metadata_provider: Callable[[Dict[str, Any]], Dict[str, Any]]
    ) -> None:
        self._metadata_provider = metadata_provider

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._inject(source)

    def _inject(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            index = 0

            def on_next(item: FLVStreamItem) -> None:
                nonlocal index
                if isinstance(item, FlvHeader):
                    index = 0
                    observer.on_next(item)
                    return

                tag = item
                index += 1

                if index == 1:
                    if is_metadata_tag(tag):
                        tag = self._inject_metadata(tag)
                        logger.debug('Injected metadata into the metadata tag')
                    else:
                        logger.debug('No metadata tag in the stream')
                        metadata_tag = self._make_metadata_tag()
                        logger.debug('Maked a metadata tag for metadata injection')
                        observer.on_next(metadata_tag)
                        logger.debug('Inserted the artificial metadata tag')

                observer.on_next(tag)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _inject_metadata(self, tag: ScriptTag) -> ScriptTag:
        old_metadata = parse_metadata(tag)
        new_metadata = self._metadata_provider(old_metadata.copy())
        final_metadata = {
            **{'duration': 0.0, 'filesize': 0.0},
            **old_metadata,
            **new_metadata,
        }
        new_tag = enrich_metadata(tag, final_metadata, offset=tag.offset)

        if 'keyframes' in final_metadata:
            keyframes = cast(KeyFramesDict, final_metadata['keyframes'])
            offset = new_tag.tag_size - tag.tag_size
            keyframes['filepositions'] = list(
                map(lambda p: p + offset, keyframes['filepositions'])
            )
            if 'lastkeyframelocation' in final_metadata:
                final_metadata['lastkeyframelocation'] = keyframes['filepositions'][-1]
            new_tag = update_metadata(new_tag, final_metadata)

        return new_tag

    def _make_metadata_tag(self) -> ScriptTag:
        metadata = self._metadata_provider({})
        metadata = {'duration': 0.0, 'filesize': 0.0, **metadata}
        return create_metadata_tag(metadata)




import signal
from typing import Any, Dict

import attr
from fastapi import APIRouter, status

from ..schemas import ResponseMessage
from ...application import Application


app: Application = None  # type: ignore  # bypass flake8 F821

router = APIRouter(
    prefix='/api/v1/app',
    tags=['application'],
)


@router.get(
    '/status',
)
async def get_app_status() -> Dict[str, Any]:
    return attr.asdict(app.status)


@router.get(
    '/info',
)
async def get_app_info() -> Dict[str, Any]:
    return attr.asdict(app.info)


@router.post(
    '/restart',
    response_model=ResponseMessage,
)
async def restart_app() -> ResponseMessage:
    await app.restart()
    return ResponseMessage(message='Application has been restarted')


@router.post(
    '/exit',
    status_code=status.HTTP_204_NO_CONTENT,
)
async def exit_app() -> None:
    signal.raise_signal(signal.SIGINT)



from __future__ import annotations

import io
import logging
from typing import List, Optional, Union

from reactivex import Observable, Subject, abc

from ...utils.ffprobe import StreamProfile, ffprobe
from .segment_fetcher import InitSectionData, SegmentData

__all__ = ('HLSProber', 'StreamProfile')


logger = logging.getLogger(__name__)


class HLSProber:
    def __init__(self) -> None:
        self._profiles: Subject[StreamProfile] = Subject()

    def _reset(self) -> None:
        self._gathering: bool = False
        self._gathered_items: List[Union[InitSectionData, SegmentData]] = []

    @property
    def profiles(self) -> Observable[StreamProfile]:
        return self._profiles

    def __call__(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        return self._probe(source)

    def _probe(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        def subscribe(
            observer: abc.ObserverBase[Union[InitSectionData, SegmentData]],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            self._reset()

            def on_next(item: Union[InitSectionData, SegmentData]) -> None:
                if isinstance(item, InitSectionData):
                    self._gathered_items.clear()
                    self._gathering = True

                if self._gathering:
                    self._gathered_items.append(item)
                    if len(self._gathered_items) >= 10:
                        try:
                            self._do_probe()
                        except Exception as e:
                            logger.warning(f'Failed to probe stream: {repr(e)}')
                        finally:
                            self._gathered_items.clear()
                            self._gathering = False

                observer.on_next(item)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _do_probe(self) -> None:
        bytes_io = io.BytesIO()
        for item in self._gathered_items:
            bytes_io.write(item.payload)

        def on_next(profile: StreamProfile) -> None:
            self._profiles.on_next(profile)

        def on_error(e: Exception) -> None:
            logger.warning(f'Failed to probe stream by ffprobe: {repr(e)}')

        ffprobe(bytes_io.getvalue()).subscribe(on_next, on_error)



import logging
import os
from pathlib import Path
from typing import Iterator, Optional

from ..bili.danmaku_client import DanmakuClient
from ..bili.live import Live
from ..bili.live_monitor import LiveMonitor
from ..bili.models import RoomInfo, UserInfo
from ..bili.typing import QualityNumber, StreamFormat
from ..core import Recorder
from ..core.cover_downloader import CoverSaveStrategy
from ..event.event_submitters import (
    LiveEventSubmitter,
    PostprocessorEventSubmitter,
    RecorderEventSubmitter,
)
from ..flv.metadata_injection import InjectingProgress
from ..flv.operators import MetaData, StreamProfile
from ..logging.room_id import aio_task_with_room_id
from ..postprocess import DeleteStrategy, Postprocessor, PostprocessorStatus
from ..postprocess.remux import RemuxingProgress
from .models import (
    DanmakuFileDetail,
    DanmukuFileStatus,
    RunningStatus,
    TaskStatus,
    VideoFileDetail,
    VideoFileStatus,
)

__all__ = ('RecordTask',)


logger = logging.getLogger(__name__)


class RecordTask:
    def __init__(
        self,
        room_id: int,
        *,
        out_dir: str = '',
        path_template: str = '',
        cookie: str = '',
        user_agent: str = '',
        remux_to_mp4: bool = False,
        inject_extra_metadata: bool = True,
        delete_source: DeleteStrategy = DeleteStrategy.AUTO,
    ) -> None:
        super().__init__()

        self._live = Live(room_id, user_agent, cookie)

        self._room_id = room_id
        self._out_dir = out_dir
        self._path_template = path_template
        self._cookie = cookie
        self._user_agent = user_agent
        self._remux_to_mp4 = remux_to_mp4
        self._inject_extra_metadata = inject_extra_metadata
        self._delete_source = delete_source

        self._ready = False
        self._monitor_enabled: bool = False
        self._recorder_enabled: bool = False

    @property
    def ready(self) -> bool:
        return self._ready

    @property
    def room_info(self) -> RoomInfo:
        return self._live.room_info

    @property
    def user_info(self) -> UserInfo:
        return self._live.user_info

    @property
    def running_status(self) -> RunningStatus:
        if not self._monitor_enabled and not self._recorder_enabled:
            return RunningStatus.STOPPED
        elif self._recorder.recording:
            return RunningStatus.RECORDING
        elif self._postprocessor.status == PostprocessorStatus.REMUXING:
            return RunningStatus.REMUXING
        elif self._postprocessor.status == PostprocessorStatus.INJECTING:
            return RunningStatus.INJECTING
        else:
            return RunningStatus.WAITING

    @property
    def monitor_enabled(self) -> bool:
        return self._monitor_enabled

    @property
    def recorder_enabled(self) -> bool:
        return self._recorder_enabled

    @property
    def status(self) -> TaskStatus:
        return TaskStatus(
            monitor_enabled=self.monitor_enabled,
            recorder_enabled=self.recorder_enabled,
            running_status=self.running_status,
            stream_url=self._recorder.stream_url,
            stream_host=self._recorder.stream_host,
            dl_total=self._recorder.dl_total,
            dl_rate=self._recorder.dl_rate,
            rec_elapsed=self._recorder.rec_elapsed,
            rec_total=self._recorder.rec_total,
            rec_rate=self._recorder.rec_rate,
            danmu_total=self._recorder.danmu_total,
            danmu_rate=self._recorder.danmu_rate,
            real_stream_format=self._recorder.real_stream_format,
            real_quality_number=self._recorder.real_quality_number,
            recording_path=self.recording_path,
            postprocessor_status=self._postprocessor.status,
            postprocessing_path=self._postprocessor.postprocessing_path,
            postprocessing_progress=(self._postprocessor.postprocessing_progress),
        )

    @property
    def video_file_details(self) -> Iterator[VideoFileDetail]:
        recording_paths = set(self._recorder.get_recording_files())
        completed_paths = set(self._postprocessor.get_completed_files())

        for path in self._recorder.get_video_files():
            try:
                size = os.path.getsize(path)
                exists = True
            except FileNotFoundError:
                mp4_path = str(Path(path).with_suffix('.mp4'))
                try:
                    size = os.path.getsize(mp4_path)
                    exists = True
                    path = mp4_path
                except FileNotFoundError:
                    size = 0
                    exists = False

            if not exists:
                status = VideoFileStatus.MISSING
            elif path in completed_paths:
                status = VideoFileStatus.COMPLETED
            elif path in recording_paths:
                status = VideoFileStatus.RECORDING
            elif path == self._postprocessor.postprocessing_path:
                progress = self._postprocessor.postprocessing_progress
                if isinstance(progress, RemuxingProgress):
                    status = VideoFileStatus.REMUXING
                elif isinstance(progress, InjectingProgress):
                    status = VideoFileStatus.INJECTING
                else:
                    if self._postprocessor.remux_to_mp4:
                        status = VideoFileStatus.REMUXING
                    else:
                        status = VideoFileStatus.INJECTING
            else:
                # disabling recorder by force or stoping task by force
                status = VideoFileStatus.BROKEN

            yield VideoFileDetail(path=path, size=size, status=status)

    @property
    def danmaku_file_details(self) -> Iterator[DanmakuFileDetail]:
        recording_paths = set(self._recorder.get_recording_files())
        completed_paths = set(self._postprocessor.get_completed_files())

        for path in self._recorder.get_danmaku_files():
            try:
                size = os.path.getsize(path)
                exists = True
            except FileNotFoundError:
                size = 0
                exists = False

            if not exists:
                status = DanmukuFileStatus.MISSING
            elif path in completed_paths:
                status = DanmukuFileStatus.COMPLETED
            elif path in recording_paths:
                status = DanmukuFileStatus.RECORDING
            else:
                # disabling recorder by force or stoping task by force
                status = DanmukuFileStatus.BROKEN

            yield DanmakuFileDetail(path=path, size=size, status=status)

    @property
    def user_agent(self) -> str:
        return self._live.user_agent

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        self._live.user_agent = value

    @property
    def cookie(self) -> str:
        return self._live.cookie

    @cookie.setter
    def cookie(self, value: str) -> None:
        self._live.cookie = value

    @property
    def danmu_uname(self) -> bool:
        return self._recorder.danmu_uname

    @danmu_uname.setter
    def danmu_uname(self, value: bool) -> None:
        self._recorder.danmu_uname = value

    @property
    def record_gift_send(self) -> bool:
        return self._recorder.record_gift_send

    @record_gift_send.setter
    def record_gift_send(self, value: bool) -> None:
        self._recorder.record_gift_send = value

    @property
    def record_free_gifts(self) -> bool:
        return self._recorder.record_free_gifts

    @record_free_gifts.setter
    def record_free_gifts(self, value: bool) -> None:
        self._recorder.record_free_gifts = value

    @property
    def record_guard_buy(self) -> bool:
        return self._recorder.record_guard_buy

    @record_guard_buy.setter
    def record_guard_buy(self, value: bool) -> None:
        self._recorder.record_guard_buy = value

    @property
    def record_super_chat(self) -> bool:
        return self._recorder.record_super_chat

    @record_super_chat.setter
    def record_super_chat(self, value: bool) -> None:
        self._recorder.record_super_chat = value

    @property
    def save_cover(self) -> bool:
        return self._recorder.save_cover

    @save_cover.setter
    def save_cover(self, value: bool) -> None:
        self._recorder.save_cover = value

    @property
    def cover_save_strategy(self) -> CoverSaveStrategy:
        return self._recorder.cover_save_strategy

    @cover_save_strategy.setter
    def cover_save_strategy(self, value: CoverSaveStrategy) -> None:
        self._recorder.cover_save_strategy = value

    @property
    def save_raw_danmaku(self) -> bool:
        return self._recorder.save_raw_danmaku

    @save_raw_danmaku.setter
    def save_raw_danmaku(self, value: bool) -> None:
        self._recorder.save_raw_danmaku = value

    @property
    def stream_format(self) -> StreamFormat:
        return self._recorder.stream_format

    @stream_format.setter
    def stream_format(self, value: StreamFormat) -> None:
        self._recorder.stream_format = value

    @property
    def quality_number(self) -> QualityNumber:
        return self._recorder.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._recorder.quality_number = value

    @property
    def fmp4_stream_timeout(self) -> int:
        return self._recorder.fmp4_stream_timeout

    @fmp4_stream_timeout.setter
    def fmp4_stream_timeout(self, value: int) -> None:
        self._recorder.fmp4_stream_timeout = value

    @property
    def real_stream_format(self) -> Optional[StreamFormat]:
        return self._recorder.real_stream_format

    @property
    def real_quality_number(self) -> Optional[QualityNumber]:
        return self._recorder.real_quality_number

    @property
    def buffer_size(self) -> int:
        return self._recorder.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._recorder.buffer_size = value

    @property
    def read_timeout(self) -> int:
        return self._recorder.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._recorder.read_timeout = value

    @property
    def disconnection_timeout(self) -> int:
        return self._recorder.disconnection_timeout

    @disconnection_timeout.setter
    def disconnection_timeout(self, value: int) -> None:
        self._recorder.disconnection_timeout = value

    @property
    def out_dir(self) -> str:
        return self._recorder.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._recorder.out_dir = value

    @property
    def path_template(self) -> str:
        return self._recorder.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._recorder.path_template = value

    @property
    def filesize_limit(self) -> int:
        return self._recorder.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._recorder.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._recorder.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._recorder.duration_limit = value

    @property
    def recording_path(self) -> Optional[str]:
        return self._recorder.recording_path

    @property
    def metadata(self) -> Optional[MetaData]:
        return self._recorder.metadata

    @property
    def stream_profile(self) -> StreamProfile:
        return self._recorder.stream_profile

    @property
    def remux_to_mp4(self) -> bool:
        return self._postprocessor.remux_to_mp4

    @remux_to_mp4.setter
    def remux_to_mp4(self, value: bool) -> None:
        self._postprocessor.remux_to_mp4 = value

    @property
    def inject_extra_metadata(self) -> bool:
        return self._postprocessor.inject_extra_metadata

    @inject_extra_metadata.setter
    def inject_extra_metadata(self, value: bool) -> None:
        self._postprocessor.inject_extra_metadata = value

    @property
    def delete_source(self) -> DeleteStrategy:
        return self._postprocessor.delete_source

    @delete_source.setter
    def delete_source(self, value: DeleteStrategy) -> None:
        self._postprocessor.delete_source = value

    def can_cut_stream(self) -> bool:
        return self._recorder.can_cut_stream()

    def cut_stream(self) -> bool:
        return self._recorder.cut_stream()

    @aio_task_with_room_id
    async def setup(self) -> None:
        await self._live.init()
        await self._setup()
        self._ready = True

    @aio_task_with_room_id
    async def destroy(self) -> None:
        await self._destroy()
        await self._live.deinit()
        self._ready = False

    @aio_task_with_room_id
    async def enable_monitor(self) -> None:
        if self._monitor_enabled:
            return
        self._monitor_enabled = True

        await self._danmaku_client.start()
        self._live_monitor.enable()

    @aio_task_with_room_id
    async def disable_monitor(self) -> None:
        if not self._monitor_enabled:
            return
        self._monitor_enabled = False

        self._live_monitor.disable()
        await self._danmaku_client.stop()

    @aio_task_with_room_id
    async def enable_recorder(self) -> None:
        if self._recorder_enabled:
            return
        self._recorder_enabled = True

        await self._postprocessor.start()
        await self._recorder.start()

    @aio_task_with_room_id
    async def disable_recorder(self, force: bool = False) -> None:
        if not self._recorder_enabled:
            return
        self._recorder_enabled = False

        if force:
            await self._postprocessor.stop()
            await self._recorder.stop()
        else:
            await self._recorder.stop()
            await self._postprocessor.stop()

    async def update_info(self, raise_exception: bool = False) -> bool:
        return await self._live.update_info(raise_exception=raise_exception)

    @aio_task_with_room_id
    async def update_session(self) -> None:
        if self._monitor_enabled:
            await self._danmaku_client.stop()

        await self._live.deinit()
        await self._live.init()
        self._danmaku_client.session = self._live.session
        self._danmaku_client.appapi = self._live.appapi
        self._danmaku_client.webapi = self._live.webapi

        if self._monitor_enabled:
            await self._danmaku_client.start()

    async def _setup(self) -> None:
        self._setup_danmaku_client()
        self._setup_live_monitor()
        self._setup_live_event_submitter()
        self._setup_recorder()
        self._setup_recorder_event_submitter()
        self._setup_postprocessor()
        self._setup_postprocessor_event_submitter()

    def _setup_danmaku_client(self) -> None:
        self._danmaku_client = DanmakuClient(
            self._live.session, self._live.appapi, self._live.webapi, self._live.room_id
        )

    def _setup_live_monitor(self) -> None:
        self._live_monitor = LiveMonitor(self._danmaku_client, self._live)

    def _setup_live_event_submitter(self) -> None:
        self._live_event_submitter = LiveEventSubmitter(self._live_monitor)

    def _setup_recorder(self) -> None:
        self._recorder = Recorder(
            self._live,
            self._danmaku_client,
            self._live_monitor,
            self._out_dir,
            self._path_template,
        )

    def _setup_recorder_event_submitter(self) -> None:
        self._recorder_event_submitter = RecorderEventSubmitter(self._recorder)

    def _setup_postprocessor(self) -> None:
        self._postprocessor = Postprocessor(
            self._live,
            self._recorder,
            remux_to_mp4=self._remux_to_mp4,
            inject_extra_metadata=self._inject_extra_metadata,
            delete_source=self._delete_source,
        )

    def _setup_postprocessor_event_submitter(self) -> None:
        self._postprocessor_event_submitter = PostprocessorEventSubmitter(
            self._postprocessor
        )

    async def _destroy(self) -> None:
        self._destroy_postprocessor_event_submitter()
        self._destroy_postprocessor()
        self._destroy_recorder_event_submitter()
        self._destroy_recorder()
        self._destroy_live_event_submitter()
        self._destroy_live_monitor()
        self._destroy_danmaku_client()

    def _destroy_danmaku_client(self) -> None:
        del self._danmaku_client

    def _destroy_live_monitor(self) -> None:
        del self._live_monitor

    def _destroy_live_event_submitter(self) -> None:
        del self._live_event_submitter

    def _destroy_recorder(self) -> None:
        del self._recorder

    def _destroy_recorder_event_submitter(self) -> None:
        del self._recorder_event_submitter

    def _destroy_postprocessor(self) -> None:
        del self._postprocessor

    def _destroy_postprocessor_event_submitter(self) -> None:
        del self._postprocessor_event_submitter
